/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        dmatest.h

    Purpose:

        This file contains the declarations for structures used in 
        dmatest.c
--*/

#ifdef ACIT_CFG_DMA_TEST

//
// The following DMA tests are done.
//
// 1. 2-way split of every SGE/Constant split size
// 2. 2-way split of every SGE/Incrementing split size
//
// 3. 3-way split of every SGE/Constant split size
// 4. 3-way split of every SGE/Incrementing split size
//
//
// - Code automatically switches from one test to another after running a million IOs
//   on each test
//
// - The split is done in incremental amounts upto 4K-1 in round robin for every SGE.
//   For e.g. the first SGE in 2-way split will chop the first SGE into 1byte and
//   Len-1 byte, the second SGE into 2bytes and Len-2 bytes, the third SGE into 3bytes
//   and Len-3 bytes and so on.
//   
// - For every N-way split, consecutive PAs are always placed in consecutive SGE locations
//   to ensure that the SGE offsets are incrementing linearly
//
// - Because of a N-way split of every SGE, the number of ISCSI_SGE allocated per SGL will
//   not be enough. Hence care has to be taken to configure the workloads to not cross this
//   allocation. One option is to enable ACIT_CFG_CHAINED_SGL so that the driver will
//   chain SGLs if ISCSI_SGEs in a SGL don't suffice.
//   If the tests have to be done without chaining, then large_io=512; can be enabled and
//   the application be configured to try a max of 332K (works out to 250SGEs/SGL). To be
//   safe, a max of 256K can be tried.
//

#define DMA_TEST_SPLIT_SZ_CONSTANT      (4096)
#define DMA_TEST_ROLL_OVER_LIMIT        (100000)
#define DMA_TEST_INC_FRAG_NUM_AT        (10000)

// Split level
#define DMA_TEST_SPLIT_2WAY             (2)
#define DMA_TEST_SPLIT_3WAY             (3)

static const ACIT_UEXACT8 dmatest_str[][64] = {
    "Disabled",
    "Enabled",
    "Split SGE : 3-Way Constant Split",
    "Split SGE : 3-Way Incremental Split",
    "Split SGE : 2-Way Constant Split",
    "Split SGE : 2-Way Incremental Split",
    "FragNum Offset : Increment FragNum at rollover",
    "FragNum Offset : Increment FragNum per IO",
    "FragNum Offset : Chaining within a ISCSI_SGL",
};

// DMA test cases
#define DMA_TEST_DISABLED                (0)
#define DMA_TEST_ENABLED                 (1)
#define DMA_TEST_3WAY_CONST_SPLIT        (2)
#define DMA_TEST_3WAY_INCRE_SPLIT        (3)
#define DMA_TEST_2WAY_CONST_SPLIT        (4)
#define DMA_TEST_2WAY_INCRE_SPLIT        (5)
#define DMA_TEST_INC_FRAG_NUM            (6)
#define DMA_TEST_SEQINC_FRAG_NUM         (7)
#define DMA_TEST_MULTIPLE_CHAINS_PER_SGL (8)

#define DMA_TEST_START_TEST_CASE         (2)
#define DMA_TEST_MAX_TEST_CASE           (8)

// Max starting offset of frag_num in a 32 SGE element array
#define DMA_TEST_FRAG_NUM_MAX_START_OFFSET (30)

#define FILL_ACIT_DMAT_SGE(sge, _a64, _len) {   \
    sge.address.u.a64.address = _a64;           \
    sge.length = _len;                          \
}

ACIT_VOID
ap_init_dma_test(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
ap_init_dmatest_sg_list(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request,
    PACIT_IO_WS             pio);

ACIT_VOID
ap_dma_test_nway_split(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request,
    PACIT_IO_WS             pio,
    ACIT_UEXACT32           const_split_sz,
    ACIT_UEXACT32           n);

ACIT_VOID
ap_dma_test_split_64k_sge(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request,
    PACIT_IO_WS             pio);

ACIT_VOID
ap_check_and_inc_dma_test(
    PACIT_CONTROLLER_WS     pcontroller);

#pragma optimize( "", off )
STATIC ACIT_VOID
ap_crash_system(PACIT_UEXACT32 ptr)
{
    *ptr = 0;
}
#pragma optimize( "", on ) 

#endif
