/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!

@file
    cq.cpp

@brief
    Implementation for CQ objects. These objects represent and manage on-chip
    Completion Queue structures.

@author jflores

@note
    jflores   8/03/2004    initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/cq_ll.c,v 1.23 2006/04/09 23:32:54 menz Exp $

*/

#include "pch.h"

// LANCER TODO: Remove once LANCER specific defs find their right place
#include "lancer_temp.h"

/*!
@brief
    This routines searches for a CqId that is bound to a specified function object.
@param    
    function_object  - function object the CQ should be bound to.
@param    
    cq_id            - cq_id to find
@return   
    A pointer to the CQ_OBJECT that represents this CqId. NULL if not found
@note
    IRQL: <= DISPATCH_LEVEL
*/
STATIC
INLINE
PBE_CQ_OBJECT
be_cq_id_to_object (
    PBE_FUNCTION_OBJECT  function_object, 
    u32                  cq_id)
{
    PBE_CQ_OBJECT   cq_cur;
    PBE_CQ_OBJECT   cq_ret = NULL;

    _be_function_lock (function_object);

    SA_FOR_EACH_LIST_ENTRY (cq_cur, function_object->links.cq_list_head, BE_CQ_OBJECT, cq_list)
    {
        if (be_cq_get_id (cq_cur) == cq_id) {

            cq_ret = cq_cur;
            break;
        }
    }

    _be_function_unlock (function_object);

    if (!cq_ret) {
        TRACE (DL_ERR, "Failed to locate cq_id:%d for processing.", cq_id);
    }

    return  cq_ret;
}


//============================================================================
//                  P U B L I C  R O U T I N E S
//============================================================================


void
be_cq_create_callback (
    PVOID Context,
    BESTATUS Status,
    MCC_WRB* OptionalWrb)
{
    PBE_CQ_OBJECT cq_object = (PBE_CQ_OBJECT)Context;
    IOCTL_COMMON_CQ_CREATE *ioctl = (IOCTL_COMMON_CQ_CREATE*)&OptionalWrb->payload;
        
    if (Status == BE_SUCCESS && OptionalWrb) {

        // Remember the CQ id.
        cq_object->cq_id = ioctl->params.response.cq_id;

        // insert this cq into eq_object reference
        if (cq_object->eq_object) {
            _be_eq_add_cq (cq_object->eq_object, cq_object);
        }

        // insert this cq into function_object
        _be_function_add_cq (cq_object->parent_function, cq_object);

        TRACE (DL_INFO, "cq created. function:%d pd:%d cqid:%d",
               be_function_get_function_number (cq_object->parent_function),
               be_function_get_pd_number (cq_object->parent_function),
               cq_object->cq_id
               );
    }
}


BESTATUS
be_cq_create_ex (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 length,
   IN CQ_CREATE_PARAMS    *param,
   IN PBE_EQ_OBJECT       eq_object,
   OUT PBE_CQ_OBJECT      cq_object,   
   IN MCC_WRB_CQE_CALLBACK     callback,
   IN PVOID                    callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT *queue_context
)
{
    BESTATUS        status = BE_SUCCESS;
    u32             num_entries_encoding;
    u32             num_entries = length / 16; // all CQEs are same size
    IOCTL_COMMON_CQ_CREATE *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BE_MCC_WRB_RESPONSE_COPY response_copy = {0};
    
    ASSERT (sgl);
    ASSERT (cq_object);
    ASSERT (length % 16 == 0);

    FUNCTION_ASSERT (function_object);        
    
    // LANCER doesn't support version 0 of IOCTL_COMMON_CQ_CREATE command.
    // So instead version 2 of the same command should be used.
    ASSERT (function_object->adapter_family != ADAPTER_FAMILY_LANCER);

    switch (num_entries) {
        case 256: num_entries_encoding  = CEV_CQ_CNT_256; break;
        case 512: num_entries_encoding  = CEV_CQ_CNT_512; break;
        case 1024: num_entries_encoding = CEV_CQ_CNT_1024; break;
        default:
            SA_BREAKPOINT_MSG ("bad cq num entries");
            return BE_STATUS_INVALID_PARAMETER;
    }

    // All cq entries all the same size.  Use iSCSI version as a test for the proper
    // sgl length.
    sa_zero_mem (cq_object, sizeof (*cq_object));

    cq_object->magic           = BE_CQ_MAGIC;
    cq_object->ref_count       = 0;
    cq_object->parent_function = function_object;    
    cq_object->eq_object       = eq_object;    
    cq_object->num_entries     = num_entries;
    cq_object->va              = sa_sgl_get_base_va (sgl); // save for MCC cq processing

    // map into UT.
    length = num_entries * 16; // all CQ entries are same size

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB; 
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_CQ_CREATE);

    ioctl->params.request.num_pages = (UEXACT16) SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);
    
    // context
    ioctl->params.request.context.valid     = TRUE;
    ioctl->params.request.context.Func      = be_function_get_function_number (function_object);
    ioctl->params.request.context.Eventable = (eq_object != NULL);
    ioctl->params.request.context.Armed     = TRUE;
    ioctl->params.request.context.EQID      = eq_object ? eq_object->eq_id : 0;
    ioctl->params.request.context.Count     = num_entries_encoding;
    ioctl->params.request.context.PD        = be_function_get_pd_number (function_object);
    ioctl->params.request.context.NoDelay   = param->no_delay;
    ioctl->params.request.context.SolEvent  = param->solicited_eventable;            
    ioctl->params.request.context.CoalescWM = param->cqe_dma_coalescing; // See BE_CQ_DMA_COALESCING enum
    ASSERT (ioctl->params.request.context.CoalescWM == param->cqe_dma_coalescing); // Detect truncation
    
    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = // be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
             be_function_post_mcc_wrb_complete (function_object, 
                                              wrb, 
                                              generic_context,  // queue context
                                              callback, 
                                              callback_context, 
                                              be_cq_create_callback, // internal callback
                                              cq_object,             // internal callback context
                                              ioctl, 
                                              response_copy);
    if (status < 0) {
        TRACE (DL_ERR, "MCC to create CQ failed.");
        goto Error;
    }    

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


//---------------------------------------------------------------------------------------
// Function: be_cq_create
//
//   Creates a CQ. This uses a synchronous IOCTL.
//
// function_object     - 
// sgl                 - No virtual address required, except for MCC CQ.
// length              - Length is bytes of queue. Number of entries is power of 2, from
//                       256 to 1k. 
// solicited_eventable - If TRUE, only CQEs with solicited event bit cause EQE write. 
// no_delay            - TRUE means force interrupt immediately without waiting for any EQ
//                       delay. 
// cqe_dma_coalescing  - Each CQ supports coalescing the DMA of CQs into larger transfers.
//                       The maximum DMA transfer is up to 4 CQEs in a single DMA. When
//                       the CQ ring contains 'N' entries that the driver has not
//                       consumed, it will start coalescing future DMAs. The future DMA
//                       will not occur until the driver rings the CQ doorbell to reduce
//                       the outstanding number of CQ entries to less than 'N' -- or when
//                       CEV has buffered 4 CQEs to transfer. 
// 
//                       The value 'N' is
//                       configurable by this setting using the enumeration
//                       BE_CQ_DMA_COALESCING. Using a small value (i.e. 2) will allow
//                       more DMA coalescing. However, the driver must be careful about
//                       increased latencies. 
// eq_object           - Optional, if set this is an eventable CQ. 
// cq_object           - Created CQ object. 
// return status       - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 length,
   IN boolean             solicited_eventable,
   IN boolean             no_delay,
   IN u32                 cqe_dma_coalescing,
   IN PBE_EQ_OBJECT       eq_object,
   OUT PBE_CQ_OBJECT      cq_object
) 
{
    CQ_CREATE_PARAMS param = {{{0}}};
    
    param.solicited_eventable = solicited_eventable;
    param.no_delay = no_delay;
    param.cqe_dma_coalescing = cqe_dma_coalescing;
    
    return be_cq_create_ex (
        function_object,
        sgl,
        length,
        &param,    
        eq_object,
        cq_object,
        NULL, // queue context
        NULL, // callback
        NULL); // context
}


//---------------------------------------------------------------------------------------
// Function: be_cq_create_v2
//
//   Creates a CQ (Version 2). This uses a synchronous IOCTL.
//
// function_object     - 
// sgl                 - No virtual address required, except for MCC CQ.
// length              - Length is bytes of queue. Number of entries is power of 2.
// no_delay            - TRUE means force interrupt immediately without waiting for any EQ
//                       delay. 
// cqe_dma_coalescing  - Each CQ supports coalescing the DMA of CQs into larger transfers.
//                       The maximum DMA transfer is up to 4 CQEs in a single DMA. When
//                       the CQ ring contains 'N' entries that the driver has not
//                       consumed, it will start coalescing future DMAs. The future DMA
//                       will not occur until the driver rings the CQ doorbell to reduce
//                       the outstanding number of CQ entries to less than 'N' -- or when
//                       CEV has buffered 4 CQEs to transfer. 
// 
//                       The value 'N' is
//                       configurable by this setting using the enumeration
//                       BE_CQ_DMA_COALESCING. Using a small value (i.e. 2) will allow
//                       more DMA coalescing. However, the driver must be careful about
//                       increased latencies. 
// eq_object           - Optional, if set this is an eventable CQ. 
// cq_object           - Created CQ object. 
// cq_type             - Type of CQ:
//                       0x00 - Default CQ. The "normal" CQ for use with Protocol-specific WQs and RQs
//                       0x01 - DPP CQ. Used in RoCE.
// auto_valid          - This field is RESERVED field for BE3-R-based products. This
//                       field is allowed only for RDMA CQs in Lancer-R-based products.
//
//                       Indicates that the "VALID" bit in all the CQEs for this CQ will
//                       be read-only for the SLI-Host and will be updated only by the
//                       SLI-Port. In any pass through the CQ, all CQEs will have the
//                       same value for the "VALID" bit (e.g., set to 1) in all CQEs. On
//                       wrapping around for the next pass through the CQ, the "VALID"
//                       bit in all CQEs will be toggled (i.e., set to 0).
//
// return status       - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cq_create_v2 (
   IN PBE_FUNCTION_OBJECT      function_object,
   IN PSA_SGL                  sgl,
   IN u32                      length,
   IN boolean                  no_delay,
   IN u32                      cqe_dma_coalescing,
   IN PBE_EQ_OBJECT            eq_object,
   OUT PBE_CQ_OBJECT           cq_object,
   IN u8                       cq_type,
   IN u8                       auto_valid,
   IN MCC_WRB_CQE_CALLBACK     callback,
   IN PVOID                    callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT *queue_context
) 
{
    BESTATUS                  status = BE_SUCCESS;
    u32                       num_entries_encoding;
    u32                       num_entries = length / 16; // all CQEs are same size except for DPP CQ
    IOCTL_COMMON_CQ_CREATE_V2_SLI *ioctl = NULL;
    MCC_WRB                   *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BE_MCC_WRB_RESPONSE_COPY  response_copy = {0};

    ASSERT (sgl);
    ASSERT (cq_object);
    FUNCTION_ASSERT (function_object);        

    // Encode number of CQEs in this CQ
    switch (num_entries) {
        case 256: num_entries_encoding  = CQ_CTX_CNT_256_SLI; break;
        case 512: num_entries_encoding  = CQ_CTX_CNT_512_SLI; break;
        case 1024: num_entries_encoding = CQ_CTX_CNT_1024_SLI; break;
        default:
            if (num_entries > 1024) {
                num_entries_encoding = CQ_CTX_CNT_EXT_SLI;
            } else {
                SA_BREAKPOINT_MSG ("bad cq num entries");
                return BE_STATUS_INVALID_PARAMETER;
            }
    }

    sa_zero_mem (cq_object, sizeof (*cq_object));
    cq_object->magic           = BE_CQ_MAGIC;
    cq_object->ref_count       = 0;
    cq_object->parent_function = function_object;    
    cq_object->eq_object       = eq_object;    
    cq_object->num_entries     = num_entries;
    cq_object->va              = sa_sgl_get_base_va (sgl); // save for MCC cq processing

    // map into UT.
    length = num_entries * 16; // all CQ entries are same size

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB; 
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_CQ_CREATE_V2_SLI);

    ioctl->header.request.version       = 2;
    ioctl->params.request.cq_type       = cq_type;
    ioctl->params.request.page_size     = (SA_PAGE_SIZE / 4096);  // Unit of 4KB
    ioctl->params.request.context.valid = TRUE;

    if (cq_type != CQ_CREATE_TYPE_DPP_SLI) {

        ASSERT (length % 16 == 0);
        ioctl->params.request.num_pages         = (UEXACT16) SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);
        ioctl->params.request.context.Eventable = (eq_object != NULL);
        ioctl->params.request.context.Count     = num_entries_encoding;
        // LANCER (RoCE) only field. Reserved for other platforms. Take care
        // of only LE as this path is invoked only by LANCER Windows drivers
        // (NIC/RoCE).
        ioctl->params.request.context.rsvd1     = auto_valid;
        ioctl->params.request.context.NoDelay   = no_delay;
        ioctl->params.request.context.CoalescWM = cqe_dma_coalescing; // See BE_CQ_DMA_COALESCING enum
        ioctl->params.request.context.Armed     = TRUE;
        ioctl->params.request.context.EQID      = eq_object ? eq_object->eq_id : 0;
        ioctl->params.request.context.CQE_COUNT = num_entries;  // Valid only if num_entries_encoding is 0x03

    } else {

        ioctl->params.request.num_pages = 1;  // Must be 1 always

    }

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       (PHYS_ADDR*)ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_complete (function_object, 
                                                wrb, 
                                                generic_context,  // queue context
                                                callback, 
                                                callback_context, 
                                                be_cq_create_callback, // internal callback
                                                cq_object,             // internal callback context
                                                ioctl, 
                                                response_copy);
    if (status < 0) {
        TRACE (DL_ERR, "MCC to create CQ V2 failed.");
        goto Error;
    }    

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


/*!

@brief
    Returns the CQ ID for CQ. This is the ID the chip references the queue as.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    CQ ID
@note
    IRQL: any

*/
u32 
be_cq_get_id (PBE_CQ_OBJECT cq_object)
{
    CQ_ASSERT (cq_object);    
    return cq_object->cq_id;
}

/*!

@brief
    Sets the current callback to be invoked when a CQ entry is made available. 
    This need only be set if the owner of this CQ does not recieve the CQ 
    completions directly. i.e. it does not service interrupts.

@param    
    cq_object            - CQ handle returned from cq_object_create.
    
@param    
    callback            - function to invoke when the CQ needs processing.
    
@param    
    context             - opaque context to pass to callback.        

@return   
    ptr to the previous callback.
    
@note
    IRQL: any

    this routine is currently not synchronized w/ the DPC that may run the CQ.
    therefore it is the callers responsibility to ensure the CQ will not
    interrupt while this routine is executing.
*/
CQ_CALLBACK 
be_cq_object_set_callback (
    PBE_CQ_OBJECT   cq_object, 
    CQ_CALLBACK     callback, 
    PVOID           context)
{
    CQ_CALLBACK old_callback;

    CQ_ASSERT (cq_object);    

    old_callback = cq_object->callback;

    cq_object->callback          = callback;
    cq_object->callback_context  = context;

    return old_callback;
}

/*!

@brief
    This routine handles CQ processing on CQs that were 'unrecognized' by upper
    level drivers.  The only CQs that should be processed by this routine are
        * MCC CQ
        * iWARP CQ
        
    And of these, only the MCC CQ will be processed in this routine. The iWARP
    CQ will be delegated further for handling by the iWARP device driver

@param
    function_object      - handle to a function object                           
@param
    cq_id                - the cq that has just completed processing                         

@return
    

@note
    IRQL: DISPATCH_LEVEL only
    
    this routine should be called immediately when a CQ from a ULD is not 
    recognized.
*/     
void
be_cq_object_delegate_processing (
    PBE_FUNCTION_OBJECT     function_object, 
    u32                     cq_id)
{
    PBE_CQ_OBJECT   cq_object;

    FUNCTION_ASSERT (function_object);
    

    cq_object = be_cq_id_to_object (function_object, cq_id);

    if (cq_object) {

        ASSERT (cq_object->callback);
        cq_object->callback (function_object, cq_object, cq_object->callback_context);

    } else {

        TRACE (DL_WARN, "cq_id %d not found for delegation", cq_id);
    }

}



/*!

@brief
    References the given object. The object is guaranteed to remain active until
    the reference count drops to zero.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: any

*/
u32
be_cq_object_reference (
    PBE_CQ_OBJECT cq_object)
{
    CQ_ASSERT (cq_object);    
    return  sa_atomic_increment (&cq_object->ref_count);
}


/*!

@brief
    Dereferences the given object. The object is guaranteed to remain active until
    the reference count drops to zero.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: any

*/
u32
be_cq_object_dereference (
    PBE_CQ_OBJECT cq_object)
{
    CQ_ASSERT (cq_object);    
    return  sa_atomic_decrement (&cq_object->ref_count);
}



void
be_cq_destroy_callback (
    PVOID Context,
    BESTATUS Status,
    MCC_WRB* OptionalWrb)
{
    PBE_CQ_OBJECT cq_object = (PBE_CQ_OBJECT)Context;
    
    // Remove reference if this is an eventable CQ.
    if (cq_object->eq_object) {
        _be_eq_remove_cq (cq_object->eq_object, cq_object);
    }

    // Remove from the function object.
    _be_function_remove_cq (cq_object->parent_function, cq_object);

    // Zero the software tracking object.
    sa_zero_mem (cq_object, sizeof (*cq_object));
    
}

/*!

@brief
    Deferences the given object. Once the object's reference count drops to
    zero, the object is destroyed and all resources that are held by this object
    are released.  The on-chip context is also destroyed along with the queue
    ID, and any mappings made into the UT.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: IRQL < DISPATCH_LEVEL
*/
BESTATUS
be_cq_destroy_async (
    PBE_CQ_OBJECT cq_object,    
    IN MCC_WRB_CQE_CALLBACK     callback,
    IN PVOID                    callback_context,
    IN BE_GENERIC_QUEUE_CONTEXT *queue_context)
{
    BESTATUS status = 0;

    CQ_ASSERT (cq_object);
    FUNCTION_ASSERT (cq_object->parent_function);

    // Nothing should reference this CQ at this point.
    ASSERT (cq_object->ref_count == 0);
    
    // Send ioctl to destroy the CQ.
    status = be_function_ring_destroy_async (cq_object->parent_function, 
                                             cq_object->cq_id, 
                                             IOCTL_RING_TYPE_CQ,
                                             queue_context,
                                             callback,
                                             callback_context,
                                             be_cq_destroy_callback,
                                             cq_object);
    return status;
    }


BESTATUS
be_cq_destroy (
    PBE_CQ_OBJECT cq_object)

{
    return be_cq_destroy_async (cq_object, NULL, NULL, NULL);
}


void
be_cq_modify_callback (
    PVOID Context,
    BESTATUS Status,
    MCC_WRB* OptionalWrb)
{
    PBE_CQ_OBJECT cq_object = (PBE_CQ_OBJECT)Context;

    if (Status == BE_SUCCESS && OptionalWrb) {

        // Change associations maintained in SW
        if (cq_object->new_eq_object) {
            _be_eq_remove_cq (cq_object->eq_object, cq_object);
            cq_object->eq_object = cq_object->new_eq_object;
            _be_eq_add_cq (cq_object->eq_object, cq_object);
        }
    }

    cq_object->new_eq_object = NULL;

    return;
}


//---------------------------------------------------------------------------------------
// Function: be_cq_modify
//
//   This command modifies the EQ associated with an existing CQ.
//
// function_object  - 
// cq_object        - Created CQ object. 
// eq_object        - Optional, if a non NULL EQ object is specified, CQ will be
//                    associated with it and association with old EQ will no longer exist.
// cwm_cntrl        - Coalesc Watermark control flags: 0: Disable watermark interrupt mode
//                    for this CQ 1: Disable watermakr interrupt mode and set watermark
//                    threshold specified in cwm_val for this CQ 2: Enable watermark
//                    interrupt mode for this CQ 3: Enable watermark interrupt mode for
//                    this CQ and set watermark threshold in cwm_val for this CQ 
// cwm_val          - Coalesc Watermark value. 
// callback         - Callback function on completion
// callback_context - Callback context
// queue_context    - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                    ring is full, this structure is used to queue the operation. It will
//                    be posted to the MCC ring when space becomes available. All queued
//                    commands will be posted to the ring in the order they are received.
//                    It is always valid to pass a pointer to a generic
//                    BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs are
//                    generally smaller than the generic struct. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cq_modify (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN PBE_CQ_OBJECT             cq_object,   
   IN PBE_EQ_OBJECT             eq_object,
   IN u8                        cwm_cntrl,
   IN u8                        cwm_val,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT  *queue_context
)
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_MODIFY_CQ_SLI *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BE_MCC_WRB_RESPONSE_COPY response_copy = {0};

    ASSERT (cq_object);
    FUNCTION_ASSERT (function_object);        

    if (cwm_cntrl > MOD_CQ_EN_WM_COALESC_VALID_SLI) {
        TRACE (DL_WARN, "modify_cq: bad watermak control flag %d", cwm_cntrl);
        return BE_STATUS_INVALID_PARAMETER;
    }

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB; 
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_MODIFY_CQ_SLI);

    ioctl->params.request.cq_id = cq_object->cq_id;
    ioctl->params.request.cwm_ctrl = cwm_cntrl;
    ioctl->params.request.coalesc_cwm = cwm_val;
    if (eq_object) {
        ioctl->params.request.eq_id_valid = 1;
        ioctl->params.request.eq_id = eq_object->eq_id;
        cq_object->new_eq_object = eq_object;
    }

    // Post the Ioctl
    status = be_function_post_mcc_wrb_complete (function_object, 
                                                wrb, 
                                                generic_context,  // queue context
                                                callback, 
                                                callback_context, 
                                                be_cq_modify_callback, // internal callback
                                                cq_object,             // internal callback context
                                                ioctl, 
                                                response_copy);

    if (status < 0) {
        TRACE (DL_ERR, "Modify CQ failed.");
        cq_object->new_eq_object = NULL;
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}
