/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

///*++
//
//    Filename:
//
//        dmatest.c
//
//    Purpose:
//
//        This file implements functionality to test the SGL chaining and DMA
//        operations. Various DMA tests are implemented. Look at the dmatest.h
//        file for details.
//        
//
//--*/
//
//
#include "acitpch.h"

#ifdef ACIT_CFG_DMA_TEST

ACIT_VOID
ap_init_dma_test(
    PACIT_CONTROLLER_WS     pcontroller)
{
    // Run each test for 5 mins
    pcontroller->dma_test.test_duration = 60*5;

    if(pcontroller->params.dma_test)
    {
        ACIT_UEXACT32 test = pcontroller->dma_test.test_case;

        //
        // If dma_test has been configured to be a value other than 1, test will
        // directly start from that test onwards. If not, then the test will
        // begin from the compile time value
        //
        if(pcontroller->params.dma_test != 1)
            test = pcontroller->params.dma_test;
        else
            test = DMA_TEST_START_TEST_CASE;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("***************************************************************************\n"));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("            Starting DMA Test [%s (%d)]\n", dmatest_str[test], test));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("***************************************************************************\n"));

        pcontroller->dma_test.test_case = test;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("***************************************************************************\n"));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("                    !!! DMA Test Disabled !!!\n"));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("***************************************************************************\n"));
    }
}

ACIT_VOID
ap_init_dmatest_sg_list(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request,
    PACIT_IO_WS             pio)
{
    ACIT_UEXACT32           i;

    switch(pcontroller->dma_test.test_case)
    {
    case DMA_TEST_2WAY_CONST_SPLIT:

        ap_dma_test_nway_split(pcontroller,
                    pio_request,
                    pio,
                    ACIT_TRUE,
                    DMA_TEST_SPLIT_2WAY);
        break;

    case DMA_TEST_2WAY_INCRE_SPLIT:

        ap_dma_test_nway_split(pcontroller,
                    pio_request,
                    pio,
                    ACIT_FALSE,
                    DMA_TEST_SPLIT_2WAY);
        break;

    case DMA_TEST_3WAY_CONST_SPLIT:

        ap_dma_test_nway_split(pcontroller,
                    pio_request,
                    pio,
                    ACIT_TRUE,
                    DMA_TEST_SPLIT_3WAY);
        break;

    case DMA_TEST_3WAY_INCRE_SPLIT:

        ap_dma_test_nway_split(pcontroller,
                    pio_request,
                    pio,
                    ACIT_FALSE,
                    DMA_TEST_SPLIT_3WAY);
        break;

    case DMA_TEST_MULTIPLE_CHAINS_PER_SGL:
    case DMA_TEST_INC_FRAG_NUM:
    case DMA_TEST_SEQINC_FRAG_NUM:

        // For all the frag_num based tests, force each SGE to <64K size
        ap_dma_test_split_64k_sge(pcontroller,
                                  pio_request,
                                  pio);

        break;
    }

    // print the newly constructed list
    if(pio_request->psg_list && pio_request->psg_list->number_elements)
    {
        for(i=0; i<pio_request->psg_list->number_elements; i++)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("SGE[%3.3d] - PA : 0x%p Len : 0x%x\n",
                i, pio_request->psg_list->sglist[i].address.u.a64.address,
                pio_request->psg_list->sglist[i].length));
        }
    }
}

ACIT_VOID
ap_dma_test_nway_split(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request,
    PACIT_IO_WS             pio,
    ACIT_UEXACT32           const_split_sz,
    ACIT_UEXACT32           n)
{
    ACIT_UEXACT32           last_split_sz;
    ACIT_UEXACT32           L[4], i, j, offset, idx;

    last_split_sz = pcontroller->dma_test.last_split_sz;

    if(last_split_sz == 0) last_split_sz = n-1;

    if(pio_request->psg_list && pio_request->psg_list->number_elements)
    {
        pio->num_sge = pio_request->psg_list->number_elements;
        pio->orig_sgl = pio_request->psg_list;

        // Setup SGList in IORequest to internally allocated SG List
        pio_request->psg_list = (PACIT_SG_LIST)&pio->sgl;

        for(i=0, j=0; i<pio->orig_sgl->number_elements; i++, j+=n)
        {
            L[0] = pio->orig_sgl->sglist[i].length;

            if(L[0] <= last_split_sz)
            {
                // If an element is last_split_sz, then split it at 1 byte always
                L[0] = pio->orig_sgl->sglist[i].length - (n-1);

                // Fill rest of the n-1 elements to 1
                for(idx=1; idx<=n-1; idx++)
                {
                    L[idx] = 1;
                }
            }
            else
            {
                ACIT_UEXACT32 remain;

                L[0] = pio->orig_sgl->sglist[i].length - last_split_sz;

                remain = last_split_sz;
                idx = 1;

                // Fill rest of the n-1 elements
                while(remain > 0)
                {
                    L[idx] = remain/(n-idx);
                    remain -= L[idx];
                    idx++;
                }
            }

            offset = 0;

            for(idx=0; idx<n; idx++)
            {
                OSM_ASSERT_MSG((L[idx] != 0), ("Index %d is 0\n", idx));
                FILL_ACIT_DMAT_SGE(pio->sgl.sge[j+idx],
                                   pio->orig_sgl->sglist[i].address.u.a64.address + offset,
                                   L[idx]);

                offset += L[idx];
            }

            OSM_ASSERT(offset == pio->orig_sgl->sglist[i].length);

            if(const_split_sz == 0)
            {
                // If constant split size is not set, increment split size after every SGEs
                last_split_sz++;
                if(last_split_sz >= DMA_TEST_SPLIT_SZ_CONSTANT)
                    last_split_sz = n-1;
            }
        }

        // Patch the number of SG elements
        pio_request->psg_list->number_elements = pio->orig_sgl->number_elements * n;

        // Update dma_test context area
        pcontroller->dma_test.last_split_sz = last_split_sz;

        if(const_split_sz)
        {
            // If constant split size, increment split size only after all SGEs
            // are filled in
            last_split_sz++;
            if(last_split_sz >= DMA_TEST_SPLIT_SZ_CONSTANT)
                last_split_sz = n-1;
        }
    }
}

ACIT_VOID
ap_dma_test_split_64k_sge(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request,
    PACIT_IO_WS             pio)
{
    ACIT_UEXACT32           i, j;

    if((pio_request->psg_list) && 
       (pio_request->psg_list->number_elements) &&
       (pio_request->data_transfer_length > BE_MAX_BYTES_PER_FRAG))
    {
        pio->num_sge = pio_request->psg_list->number_elements;
        pio->orig_sgl = pio_request->psg_list;

        // Setup SGList in IORequest to internally allocated SG List
        pio_request->psg_list = (PACIT_SG_LIST)&pio->sgl;

        for(i=0, j=0; i<pio->orig_sgl->number_elements; i++)
        {
            if(pio->orig_sgl->sglist[i].length > BE_MAX_BYTES_PER_FRAG)
            {
                ACIT_UEXACT32 offset = 0, len;
                ACIT_UEXACT32 remain = len = pio->orig_sgl->sglist[i].length;

                while(remain > 0)
                {
                    len = (remain > BE_MAX_BYTES_PER_FRAG) ? BE_MAX_BYTES_PER_FRAG : remain;
                    
                    FILL_ACIT_DMAT_SGE(pio->sgl.sge[j],
                                       pio->orig_sgl->sglist[i].address.u.a64.address + offset,
                                       len);

                    remain -= len;
                    offset += len;
                    j++;
                }
            }
            else
            {
                FILL_ACIT_DMAT_SGE(pio->sgl.sge[j],
                                   pio->orig_sgl->sglist[i].address.u.a64.address,
                                   pio->orig_sgl->sglist[i].length);

                j++;
            }
        }

        pio_request->psg_list->number_elements = j;
    }
}

ACIT_VOID
ap_check_and_inc_dma_test(
    PACIT_CONTROLLER_WS     pcontroller)
{
    ACIT_UEXACT32           roll_over = 0;

    // If the time from the last check, there are over 1K IOs, roll over to the
    // next test
    if((pcontroller->stats.ios_submitted - pcontroller->dma_test.last_io) > 1024)
        roll_over = 1;

    if(roll_over)
    {
        ACIT_UEXACT32 test = pcontroller->dma_test.test_case;

        test++;
        if(test > DMA_TEST_MAX_TEST_CASE)
        {
            if(pcontroller->params.dma_test != 1)
                test = pcontroller->params.dma_test;
            else
                test = DMA_TEST_START_TEST_CASE;
        }

        pcontroller->dma_test.test_case = test;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("***************************************************************************\n"));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("            Starting DMA Test [%s (%d)]\n", dmatest_str[test], test));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("***************************************************************************\n"));

        // Reset split_size and chain_at
        pcontroller->dma_test.last_split_sz = 0;
        pcontroller->dma_test.chain_at = 0;
    }

    pcontroller->dma_test.last_io = pcontroller->stats.ios_submitted;
}

#else

//
// Compilation note
// C4206: Translation unit cannot be empty
//
int dmatest_dummy = 1;

#endif
