/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        hwi.c

    Purpose:

        This file implements the Hardware Interface (hwi) for the ACIT library.
        Functions implemented in this library are specific to BladeEngine ASIC.
        

--*/


#include "acitpch.h"

ACIT_UEXACT32
hwi_get_ws_sizes(
    PACIT_UEXACT32          pcontroller_ws_size,
    PACIT_UEXACT32          ptarget_ws_size,
    PACIT_UEXACT32          plun_ws_size,
    PACIT_UEXACT32          pio_ws_size,
    PACIT_UEXACT32          ptmf_ws_size)
{
    *pcontroller_ws_size = sizeof(HWI_CONTROLLER_WS);
    *ptarget_ws_size = 0;
    *plun_ws_size = 0;
    *pio_ws_size = 0;
    *ptmf_ws_size = 0;

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_controller_ws(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PHWI_CONTROLLER_WS          phwi_controller;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    OSM_MEMSET(phwi_controller, 0, sizeof(HWI_CONTROLLER_WS));

    phwi_controller->pacit_ws = pcontroller;

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_init_rq_descriptors(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_PARAMETERS        params)
{
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS p = params;
    ACIT_UEXACT32 ulp, num_ulp;
    ACIT_UEXACT32 mem_default_pdu_ring_sz, mem_default_pdu_buffer_sz; 
    ACIT_UEXACT32 mem_default_pdu_handle_sz, async_pdu_context_sz;
    ACIT_UEXACT32 num_async_pdu_buf_page_sz;

    const ACIT_UEXACT32 a1 = ACIT_ALIGN_1BYTE;
    const ACIT_UEXACT32 a4k = ACIT_ALIGN_4096BYTE;

    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;

    // This routine could be called even if the mask has not been read from
    // hardware. The mask itself is pre-initialized to include all Ulps. The count
    // will determine how many will be initialized. This means that descriptor_base_0
    // need not be used for Ulp0 in some cases. For example, if num_ulp=1 and 
    // hardware_mask says only Ulp1 is valid, then descriptor_base_0 will be
    // used since that is the one initialized.
    //
    // If this routine is called after the ulp_mask is initialized based on ASIC
    // limits (those OSMs doing a 2 step initialization), then the memory
    // descriptors will be initialized accordingly.
    //

    //
    // Iterate through BE_MAX_ULP and check if the mask is set. If set, and the limit for
    // number of Ulps has not reached, then initialize the memory descriptors with a
    // valid length. If not, initialize the descriptors with 0 length.
    // This will ensure that only as many descriptors will be initialized as the num_ulp
    // that the OSM has programmed. Eventually, hwi_init_async_pdu_ctx will have to use
    // the right descriptor based on the hardware mask
    //

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Mask %x NumUlps %d\n", p->ulp_mask, p->num_ulp));

    for(ulp = 0, num_ulp = 0; ulp < BE_MAX_ULP; ulp++)
    {
        mem_default_pdu_buffer_sz = 0;
        mem_default_pdu_ring_sz = 0;
        mem_default_pdu_handle_sz = 0;
        async_pdu_context_sz = 0;

        if((p->ulp_mask & (1 << ulp)) &&
           (num_ulp < p->num_ulp))
        {
            num_ulp++;
            num_async_pdu_buf_page_sz = p->num_async_pdus * p->default_pdu_buf_sz;
            p->num_async_pdu_buf_pages = sa_ceiling(num_async_pdu_buf_page_sz, PAGE_SIZE);
            OSM_ASSERT(num_async_pdu_buf_page_sz <= p->num_async_pdu_buf_pages * PAGE_SIZE);

            p->num_async_pdu_buf_sgl_pages = sa_ceiling(p->num_async_pdus * sizeof(PHYS_ADDR), PAGE_SIZE);

            mem_default_pdu_buffer_sz = p->num_async_pdu_buf_pages*PAGE_SIZE;
            mem_default_pdu_ring_sz = p->num_async_pdu_buf_sgl_pages * PAGE_SIZE;
            mem_default_pdu_handle_sz = p->num_async_pdus * sizeof(ASYNC_PDU_HANDLE);
            async_pdu_context_sz = sizeof(HWI_ASYNC_PDU_CONTEXT) + 
                                ((p->num_async_pdus) * sizeof(HWI_ASYNC_ENTRY));
        }

        ap_set_mem_const (pc,
                          (PUEXACT8)"async_hdr0",
                          HWI_MEM_ASYNC_HEADER0_BUF+ulp,
                          mem_default_pdu_buffer_sz,
                          a4k,
                          non_cached);

        ap_set_mem_const (pc,
                          (PUEXACT8)"async_hdr0_ring",
                          HWI_MEM_ASYNC_HEADER0_RING+ulp,
                          mem_default_pdu_ring_sz,
                          a4k,
                          non_cached);

        ap_set_mem_const (pc,
                          (PUEXACT8)"async_hdr0_h",
                          HWI_MEM_ASYNC_HEADER0_HANDLE+ulp,
                          mem_default_pdu_handle_sz,
                          a1,
                          cached);

        ap_set_mem_const (pc,
                          (PUEXACT8)"def_buf0",
                          HWI_MEM_ASYNC_DATA0_BUF+ulp,
                          mem_default_pdu_buffer_sz,
                          a4k,
                          non_cached);

        ap_set_mem_const (pc,
                          (PUEXACT8)"def_buf0_ring",
                          HWI_MEM_ASYNC_DATA0_RING+ulp,
                          mem_default_pdu_ring_sz,
                          a4k,
                          non_cached);

        ap_set_mem_const (pc,
                          (PUEXACT8)"def_buf0_h", 
                          HWI_MEM_ASYNC_DATA0_HANDLE+ulp,
                          mem_default_pdu_handle_sz ,
                          a1,
                          cached);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Initializing AsyncPDUContext for Ulp %d : MemIdx %d Len %d \n", 
            ulp, HWI_MEM_ASYNC_PDU_CONTEXT0+ulp, async_pdu_context_sz));

        ap_set_mem_const (pc,
                          (PUEXACT8)"async_ctx0",
                          HWI_MEM_ASYNC_PDU_CONTEXT0+ulp,
                          async_pdu_context_sz,
                          a1,
                          cached);
    }
}

ACIT_UEXACT32
hwi_init_mailbox_descriptor_split(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_CONTROLLER_WS         pc = pcontroller;
    PACIT_PARAMETERS            p = &pc->params;
    PHWI_CONTROLLER_WS          ph = GET_HWI_CONTROLLER_WS(pc);

    ACIT_UEXACT32               num_wrbh_desc, i;
    ACIT_UEXACT32               remaining_cxns, num_cxns;
    ACIT_UEXACT32               num_cq_desc, remaining_cqs, num_cqs;

    const ACIT_UEXACT32 a1 = ACIT_ALIGN_1BYTE;
    const ACIT_UEXACT32 a4k = ACIT_ALIGN_4096BYTE;

    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;


    // ----------------------------------------------------
    // Pre-compute some of the intermediate sizes required.
    // ----------------------------------------------------
    ap_set_mem_const (pc, (PUEXACT8)"headers",    HWI_MEM_HDRS,             0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"eq",         HWI_MEM_EQ,               0,                          a4k, non_cached);

    hwi_init_rq_descriptors(pc, p);
    
    ap_set_mem_const (pc, (PUEXACT8)"mcc",        HWI_MEM_MCC,              0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_cq",     HWI_MEM_MCC_CQ,           0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"wrb_ctx",    HWI_MEM_WRB_CONTEXT,      0,                          a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"sglh",       HWI_MEM_SGLH,             0,                          a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_ctx",    HWI_MEM_MCC_CONTEXT,      0,                          a1, non_cached);

    // SGE are 32 per I/O.
    ph->wrb_bytes_per_cxn = sa_round_up (sizeof (ISCSI_WRB) * p->num_cmd_per_cxn, BE_PAGE_SIZE);
    ph->sge_bytes_per_io = sizeof(ISCSI_SGE) * p->num_sge;

    ph->wrb_mem_index = pc->acit_var_desc_index.range.wrb;
    ap_set_mem_variable (pc, (PUEXACT8)"wrb", ph->wrb_mem_index, 0, 0, a4k, non_cached);

    // Allocate HWI_MEM_FRAG
    ph->frag_mem_index = pc->acit_var_desc_index.range.frag;
    ap_set_mem_variable (pc, (PUEXACT8)"sge", ph->frag_mem_index, 0, 0, a4k, non_cached);

    // Allocate HWI_MEM_WRBH
    // 256 WRBHs belonging to 8 cxns will go in 1 mem descriptor
    ph->wrbh_mem_index = pc->acit_var_desc_index.range.wrbh;
    num_wrbh_desc = sa_ceiling(p->num_cxn, 8);
    remaining_cxns = p->num_cxn;
    num_cxns = 0;

    for(i = 0; i < num_wrbh_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"wrbh_"};
        mem_name[5]=(ACIT_UEXACT8)i+48;
        mem_name[6]=(ACIT_UEXACT8)'\0';

        num_cxns = (remaining_cxns > 8)? 8: num_cxns;
        ap_set_mem_variable (pc, mem_name, 
                             ph->wrbh_mem_index + i, 
                             0, 
                             0, a1, non_cached);
        remaining_cxns -= num_cxns;
    }

    ph->max_cxns_per_descriptor = 0;
    ph->max_sgl_per_descriptor  = 0;
    ph->max_wrbh_per_descriptor = 0;

    // Allocate HWI_MEM_CQ
    // 8 CQs will go in 1 mem descriptor
    ph->cq_mem_index = pc->acit_var_desc_index.range.wrbh;
    num_cq_desc = sa_ceiling(p->num_cq, 8);
    remaining_cqs = p->num_cq;
    num_cqs = 0;

    for(i = 0; i < num_cq_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"cq_"};
        mem_name[3]=(ACIT_UEXACT8)i+48;
        mem_name[4]=(ACIT_UEXACT8)'\0';

        num_cqs = (remaining_cqs > 8)? 8: num_cqs;
        ap_set_mem_variable (pc, mem_name, 
                             ph->cq_mem_index + i, 
                             0, 
                             0, a4k, non_cached);
        remaining_cqs -= num_cqs;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_mailbox_descriptor(
    PACIT_CONTROLLER_WS         pcontroller)
{
    PACIT_CONTROLLER_WS         pc = pcontroller;
    PACIT_PARAMETERS            p = &pc->params;
    PHWI_CONTROLLER_WS          ph = GET_HWI_CONTROLLER_WS(pc);

    ACIT_UEXACT32               mailbox_size;
    ACIT_UEXACT32               num_wrbh_desc, i;
    ACIT_UEXACT32               remaining_cxns, num_cxns;
    ACIT_UEXACT32               num_cq_desc, remaining_cqs, num_cqs;
    ACIT_UEXACT32               hwi_ctx_mem;

    const ACIT_UEXACT32 a1 = ACIT_ALIGN_1BYTE;
    const ACIT_UEXACT32 a4k = ACIT_ALIGN_4096BYTE;

    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;

    // Context memory has connection specific and msix specific
    // additions
    hwi_ctx_mem = sizeof(HWI_CONTEXT_MEMORY) + 
                  sizeof(BE_CXN_QUEUE_OBJECT) * p->num_cxn +
                  sizeof(HWI_INT_QUEUE_OBJECTS) * p->num_cq;
       

    // ----------------------------------------------------
    // Pre-compute some of the intermediate sizes required.
    // ----------------------------------------------------
    mailbox_size = sizeof(MCC_WRB) + sizeof(MCC_CQ_ENTRY);
    ap_set_mem_const (pc, (PUEXACT8)"mailbox",    HWI_MEM_MAILBOX,          mailbox_size,               a4k, non_cached);   
    ap_set_mem_const (pc, (PUEXACT8)"headers",    HWI_MEM_HDRS,             0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"eq",         HWI_MEM_EQ,               0,                          a4k, non_cached);

    hwi_init_rq_descriptors(pc, p);

    ap_set_mem_const (pc, (PUEXACT8)"mcc",        HWI_MEM_MCC,              0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_cq",     HWI_MEM_MCC_CQ,           0,                          a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"context",    HWI_MEM_ADDN_CONTEXT,     hwi_ctx_mem,                a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"wrb_ctx",    HWI_MEM_WRB_CONTEXT,      0,                          a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"sglh",       HWI_MEM_SGLH,             0,                          a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_ctx",    HWI_MEM_MCC_CONTEXT,      0,                          a1, non_cached);

    // SGE are 32 per I/O.
    ph->wrb_bytes_per_cxn = sa_round_up (sizeof (ISCSI_WRB) * p->num_cmd_per_cxn, BE_PAGE_SIZE);
    ph->sge_bytes_per_io = sizeof(ISCSI_SGE) * p->num_sge;

    ph->wrb_mem_index = pc->acit_var_desc_index.range.wrb;
    ap_set_mem_variable (pc, (PUEXACT8)"wrb", ph->wrb_mem_index, 0, 0, a4k, non_cached);

    // Allocate HWI_MEM_FRAG
    ph->frag_mem_index = pc->acit_var_desc_index.range.frag;
    ap_set_mem_variable (pc, (PUEXACT8)"sge", ph->frag_mem_index, 0, 0, a4k, non_cached);

    // Allocate HWI_MEM_WRBH
    // 256 WRBHs belonging to 8 cxns will go in 1 mem descriptor
    ph->wrbh_mem_index = pc->acit_var_desc_index.range.wrbh;
    num_wrbh_desc = sa_ceiling(p->num_cxn, 8);
    remaining_cxns = p->num_cxn;
    num_cxns = 0;

    for(i = 0; i < num_wrbh_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"wrbh_"};
        mem_name[5]=(ACIT_UEXACT8)i+48;
        mem_name[6]=(ACIT_UEXACT8)'\0';

        num_cxns = (remaining_cxns > 8)? 8: num_cxns;
        ap_set_mem_variable (pc, mem_name, 
                             ph->wrbh_mem_index + i, 
                             0, 
                             0, a1, non_cached);
        remaining_cxns -= num_cxns;
    }

    ph->max_cxns_per_descriptor = 0;
    ph->max_sgl_per_descriptor  = 0;
    ph->max_wrbh_per_descriptor = 0;

    // Allocate HWI_MEM_CQ
    // 8 CQs will go in 1 mem descriptor
    ph->cq_mem_index = pc->acit_var_desc_index.range.wrbh;
    num_cq_desc = sa_ceiling(p->num_cq, 8);
    remaining_cqs = p->num_cq;
    num_cqs = 0;

    for(i = 0; i < num_cq_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"cq_"};
        mem_name[3]=(ACIT_UEXACT8)i+48;
        mem_name[4]=(ACIT_UEXACT8)'\0';

        num_cqs = (remaining_cqs > 8)? 8: num_cqs;
        ap_set_mem_variable (pc, mem_name, 
                             ph->cq_mem_index + i, 
                             0, 
                             0, a4k, non_cached);
        remaining_cqs -= num_cqs;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_mem_descriptors_split(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    PHWI_CONTROLLER_WS      ph = GET_HWI_CONTROLLER_WS(pc);

    ACIT_UEXACT32           num_cq_page_sz, num_eq_page_sz;
    ACIT_UEXACT32           num_mcc_page_sz, num_mcc_cq_page_sz;
    ACIT_UEXACT32           num_wrbh_desc, i, max_phys = ACIT_CFG_MAX_CONTIGUOUS_MEM;

    ACIT_UEXACT32           remaining_cxns, num_cxns, max_mcc_wrbs;
    ACIT_UEXACT32           num_cq_desc, remaining_cqs, num_cqs;

    const ACIT_UEXACT32 a1 = ACIT_ALIGN_1BYTE;
    const ACIT_UEXACT32 a4k = ACIT_ALIGN_4096BYTE;

    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;

    // SGE are 32 per I/O.
    ph->wrb_bytes_per_cxn = sa_round_up (sizeof (ISCSI_WRB) * p->num_cmd_per_cxn, BE_PAGE_SIZE);
    ph->sge_bytes_per_io = sizeof(ISCSI_SGE) * p->num_sge;

    OSM_ASSERT(ph->wrb_bytes_per_cxn <= (sizeof(ISCSI_WRB) * BE_MAX_CMD_PER_CXN));

    // ----------------------------------------------------
    // Pre-compute some of the intermediate sizes required.
    // ----------------------------------------------------

    // number of completions is equal to the number of commands that can be pending
    num_cq_page_sz = p->num_cq_entries * sizeof (ISCSI_CQE);
    p->num_cq_pages = sa_ceiling(num_cq_page_sz, PAGE_SIZE);
    p->num_cq_entries = p->num_cq_pages * PAGE_SIZE/sizeof(ISCSI_CQE);

    num_eq_page_sz = p->num_eq_entries * sizeof (EQ_ENTRY);
    p->num_eq_pages = sa_ceiling(num_eq_page_sz, PAGE_SIZE);
    p->num_eq_entries = p->num_eq_pages * PAGE_SIZE/sizeof(EQ_ENTRY);

    // MCC ring is used by IOCTLs and invalidates, max can only be 8 pages
    max_mcc_wrbs = p->ioctls_per_ctrl;
    if(max_mcc_wrbs > MAX_MCC_WRBS)
        max_mcc_wrbs = MAX_MCC_WRBS;

    num_mcc_page_sz = (max_mcc_wrbs) * sizeof (MCC_WRB);
    p->num_mcc_pages = sa_ceiling(num_mcc_page_sz, PAGE_SIZE);

    if(p->num_mcc_pages > 8)
        p->num_mcc_pages = 8;

    num_mcc_cq_page_sz = p->ioctls_per_ctrl * sizeof (MCC_CQ_ENTRY);
    p->num_mcc_cq_pages = sa_ceiling(num_mcc_cq_page_sz, PAGE_SIZE);

    p->num_hdr_pages = sa_ceiling(p->num_cxn,32);
    ap_set_mem_const (pc, (PUEXACT8)"headers",    HWI_MEM_HDRS,           p->num_hdr_pages*PAGE_SIZE,         a4k, non_cached); // 32 headers per page

    ap_set_mem_const (pc, (PUEXACT8)"eq",         HWI_MEM_EQ,             p->num_eq_pages*PAGE_SIZE*p->num_cq, a4k, non_cached);

    hwi_init_rq_descriptors(pc, p);

    ap_set_mem_const (pc, (PUEXACT8)"mcc",        HWI_MEM_MCC,            p->num_mcc_pages*PAGE_SIZE,         a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_cq",     HWI_MEM_MCC_CQ,         p->num_mcc_cq_pages*PAGE_SIZE,      a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"wrb_ctx",    HWI_MEM_WRB_CONTEXT,    p->num_cxn*sizeof(HWI_WRB_CONTEXT), a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"sglh",       HWI_MEM_SGLH,           sizeof(SGL_HANDLE) * p->icds_per_controller, a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_ctx",    HWI_MEM_MCC_CONTEXT,    sizeof(BE_MCC_WRB_CONTEXT) * max_mcc_wrbs , a1, non_cached);

    OSM_ASSERT(sizeof(BE_MCC_WRB_CONTEXT) * p->ioctls_per_ctrl <= max_phys);

    // Allocate HWI_MEM_WRB
    ph->wrb_mem_index = pc->acit_var_desc_index.range.wrb;
    ap_set_mem_variable (pc, (PUEXACT8)"wrb", ph->wrb_mem_index, p->num_cxn, ph->wrb_bytes_per_cxn, a4k, non_cached);

    // Allocate HWI_MEM_FRAG
    ph->frag_mem_index = pc->acit_var_desc_index.range.frag;
    ap_set_mem_variable (pc, (PUEXACT8)"sge", ph->frag_mem_index, p->icds_per_controller, ph->sge_bytes_per_io, a4k, non_cached);

    // Allocate HWI_MEM_WRBH
    // 256 WRBHs belonging to 8 cxns will go in 1 mem descriptor
    ph->wrbh_mem_index = pc->acit_var_desc_index.range.wrbh;
    num_wrbh_desc = sa_ceiling(p->num_cxn, 8);
    remaining_cxns = p->num_cxn;
    num_cxns = p->num_cxn;

    for(i = 0; i < num_wrbh_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"wrbh_"};
        mem_name[5]=(ACIT_UEXACT8)i+48;
        mem_name[6]=(ACIT_UEXACT8)'\0';

        num_cxns = (remaining_cxns > 8)? 8: num_cxns;
        ap_set_mem_variable (pc, mem_name, 
                             ph->wrbh_mem_index + i, 
                             num_cxns * p->num_cmd_per_cxn, 
                             sizeof(WRB_HANDLE), a1, non_cached);
        remaining_cxns -= num_cxns;
    }

    ph->max_cxns_per_descriptor = (max_phys / ph->wrb_bytes_per_cxn);
    ph->max_sgl_per_descriptor  = (max_phys / (p->num_sge * sizeof(ISCSI_SGE)));
    ph->max_wrbh_per_descriptor = (8 * p->num_cmd_per_cxn);

    // Initialize CQ memory descriptors
    ph->cq_mem_index = pc->acit_var_desc_index.range.cq;
    num_cq_desc = sa_ceiling(p->num_cq, 8);
    remaining_cqs = p->num_cq;
    num_cqs = p->num_cq;

    for(i = 0; i < num_cq_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"cq_"};
        mem_name[3]=(ACIT_UEXACT8)i+48;
        mem_name[4]=(ACIT_UEXACT8)'\0';

        num_cqs = (remaining_cqs > 8)? 8: num_cqs;
        ap_set_mem_variable (pc, mem_name, 
                             ph->cq_mem_index + i, 
                             num_cqs * p->num_cq_entries, 
                             sizeof(ISCSI_CQE), a4k, non_cached);
        remaining_cqs -= num_cqs;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_mem_descriptors(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PACIT_PARAMETERS        p = &pc->params;
    PHWI_CONTROLLER_WS      ph = GET_HWI_CONTROLLER_WS(pc);

    ACIT_UEXACT32           num_cq_page_sz, num_eq_page_sz;
    ACIT_UEXACT32           num_mcc_page_sz, num_mcc_cq_page_sz;
    ACIT_UEXACT32           num_wrbh_desc, i, max_phys = ACIT_CFG_MAX_CONTIGUOUS_MEM;

    ACIT_UEXACT32           remaining_cxns, num_cxns, mailbox_size, max_mcc_wrbs;
    ACIT_UEXACT32           num_cq_desc, remaining_cqs, num_cqs;
    ACIT_UEXACT32           hwi_ctx_mem;

    const ACIT_UEXACT32 a1 = ACIT_ALIGN_1BYTE;
    const ACIT_UEXACT32 a4k = ACIT_ALIGN_4096BYTE;

    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;

    // SGE are 32 per I/O.
    ph->wrb_bytes_per_cxn = sa_round_up (sizeof (ISCSI_WRB) * p->num_cmd_per_cxn, BE_PAGE_SIZE);
    ph->sge_bytes_per_io = sizeof(ISCSI_SGE) * p->num_sge;

    OSM_ASSERT(ph->wrb_bytes_per_cxn <= (sizeof(ISCSI_WRB) * BE_MAX_CMD_PER_CXN));

    // ----------------------------------------------------
    // Pre-compute some of the intermediate sizes required.
    // ----------------------------------------------------
    mailbox_size = sizeof(MCC_WRB) + sizeof(MCC_CQ_ENTRY);

    // number of completions is equal to the number of commands that can be pending
    num_cq_page_sz = p->num_cq_entries * sizeof (ISCSI_CQE);
    p->num_cq_pages = sa_ceiling(num_cq_page_sz, PAGE_SIZE);
    p->num_cq_entries = p->num_cq_pages * PAGE_SIZE/sizeof(ISCSI_CQE);

    num_eq_page_sz = p->num_eq_entries * sizeof (EQ_ENTRY);
    p->num_eq_pages = sa_ceiling(num_eq_page_sz, PAGE_SIZE);
    p->num_eq_entries = p->num_eq_pages * PAGE_SIZE/sizeof(EQ_ENTRY);

    // MCC ring is used by IOCTLs and invalidates, max can only be 8 pages
    max_mcc_wrbs = p->ioctls_per_ctrl;
    if(max_mcc_wrbs > MAX_MCC_WRBS)
        max_mcc_wrbs = MAX_MCC_WRBS;

    num_mcc_page_sz = (max_mcc_wrbs) * sizeof (MCC_WRB);
    p->num_mcc_pages = sa_ceiling(num_mcc_page_sz, PAGE_SIZE);

    if(p->num_mcc_pages > 8)
        p->num_mcc_pages = 8;

    num_mcc_cq_page_sz = p->ioctls_per_ctrl * sizeof (MCC_CQ_ENTRY);
    p->num_mcc_cq_pages = sa_ceiling(num_mcc_cq_page_sz, PAGE_SIZE);

    // Context memory has connection specific and msix specific
    // additions
    hwi_ctx_mem = sizeof(HWI_CONTEXT_MEMORY) + 
                  sizeof(BE_CXN_QUEUE_OBJECT) * p->num_cxn +
                  sizeof(HWI_INT_QUEUE_OBJECTS) * p->num_cq;

    ap_set_mem_const (pc, (PUEXACT8)"mailbox",    HWI_MEM_MAILBOX,        mailbox_size,                        a4k, non_cached);   

    p->num_hdr_pages = sa_ceiling(p->num_cxn,32);
    ap_set_mem_const (pc, (PUEXACT8)"headers",    HWI_MEM_HDRS,           p->num_hdr_pages*PAGE_SIZE,         a4k, non_cached); // 32 headers per page

    ap_set_mem_const (pc, (PUEXACT8)"eq",         HWI_MEM_EQ,             p->num_eq_pages*PAGE_SIZE*p->num_cq, a4k, non_cached);

    hwi_init_rq_descriptors(pc, p);

    ap_set_mem_const (pc, (PUEXACT8)"mcc",        HWI_MEM_MCC,            p->num_mcc_pages*PAGE_SIZE,         a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_cq",     HWI_MEM_MCC_CQ,         p->num_mcc_cq_pages*PAGE_SIZE,      a4k, non_cached);
    ap_set_mem_const (pc, (PUEXACT8)"context",    HWI_MEM_ADDN_CONTEXT,   hwi_ctx_mem,                        a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"wrb_ctx",    HWI_MEM_WRB_CONTEXT,    p->num_cxn*sizeof(HWI_WRB_CONTEXT), a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"sglh",       HWI_MEM_SGLH,           sizeof(SGL_HANDLE) * p->icds_per_controller, a1, cached);
    ap_set_mem_const (pc, (PUEXACT8)"mcc_ctx",    HWI_MEM_MCC_CONTEXT,    sizeof(BE_MCC_WRB_CONTEXT) * max_mcc_wrbs , a1, non_cached);

    OSM_ASSERT(sizeof(BE_MCC_WRB_CONTEXT) * p->ioctls_per_ctrl <= max_phys);

    // Allocate HWI_MEM_WRB
    ph->wrb_mem_index = pc->acit_var_desc_index.range.wrb;
    ap_set_mem_variable (pc, (PUEXACT8)"wrb", ph->wrb_mem_index, p->num_cxn, ph->wrb_bytes_per_cxn, a4k, non_cached);

    // Allocate HWI_MEM_FRAG
    ph->frag_mem_index = pc->acit_var_desc_index.range.frag;
    ap_set_mem_variable (pc, (PUEXACT8)"sge", ph->frag_mem_index, p->icds_per_controller, ph->sge_bytes_per_io, a4k, non_cached);

    // Allocate HWI_MEM_WRBH
    // 256 WRBHs belonging to 8 cxns will go in 1 mem descriptor
    ph->wrbh_mem_index = pc->acit_var_desc_index.range.wrbh;
    num_wrbh_desc = sa_ceiling(p->num_cxn, 8);
    remaining_cxns = p->num_cxn;
    num_cxns = p->num_cxn;

    for(i = 0; i < num_wrbh_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"wrbh_"};
        mem_name[5]=(ACIT_UEXACT8)i+48;
        mem_name[6]=(ACIT_UEXACT8)'\0';

        num_cxns = (remaining_cxns > 8)? 8: num_cxns;
        ap_set_mem_variable (pc, mem_name, 
                             ph->wrbh_mem_index + i, 
                             num_cxns * p->num_cmd_per_cxn, 
                             sizeof(WRB_HANDLE), a1, non_cached);
        remaining_cxns -= num_cxns;
    }

    ph->max_cxns_per_descriptor = (max_phys / ph->wrb_bytes_per_cxn);
    ph->max_sgl_per_descriptor  = (max_phys / (p->num_sge * sizeof(ISCSI_SGE)));
    ph->max_wrbh_per_descriptor = (8 * p->num_cmd_per_cxn);

    // Initialize CQ memory descriptors
    ph->cq_mem_index = pc->acit_var_desc_index.range.cq;
    num_cq_desc = sa_ceiling(p->num_cq, 8);
    remaining_cqs = p->num_cq;
    num_cqs = p->num_cq;

    for(i = 0; i < num_cq_desc; i++)
    {
        ACIT_UEXACT8 mem_name[8] = {"cq_"};
        mem_name[3]=(ACIT_UEXACT8)i+48;
        mem_name[4]=(ACIT_UEXACT8)'\0';

        num_cqs = (remaining_cqs > 8)? 8: num_cqs;
        ap_set_mem_variable (pc, mem_name, 
                             ph->cq_mem_index + i, 
                             num_cqs * p->num_cq_entries, 
                             sizeof(ISCSI_CQE), a4k, non_cached);
        remaining_cqs -= num_cqs;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_U32
hwi_init_controller_step1(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PHWI_CONTEXT_MEMORY     phwi_context;
    ACIT_UEXACT32           desc_index, i;

    //
    // In step1, initialize the basic stuff only and invoke hwi_init_beclib_ex
    // with the desc_type as INIT. Only the basic hardware initialization will
    // be done and firmware limits queried.
    //
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    phwi_controller->phwic = (PHWI_CONTEXT_MEMORY) pcontroller->
        mem_descriptors[HWI_MEM_ADDN_CONTEXT].virtual_address;
    //
    // Initialize the context memory
    //

    phwi_context = phwi_controller->phwic;

    //
    // Bug 109828 : Read the UE mask and cache it. Note this happens prior to
    // checking if the POST is complete. So if there was a UE already and the
    // mask was modified, there is a possibility that we'll still miss the UE.
    //
    pcontroller->ue_detection.cached_ue0_mask = ACIT_PCICFG0_READ(pcontroller, ue_status_low_mask);
    pcontroller->ue_detection.cached_ue1_mask = ACIT_PCICFG0_READ(pcontroller, ue_status_hi_mask);

    if(hwi_init_beclib_ex(pcontroller, osm_handle, ACIT_DESC_TYPE_INIT) != ACIT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("hwi_init_beclib_ex ACIT_DESC_TYPE_INIT failed\n"));
        return ACIT_STATUS_ERROR;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_controller_step2(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PHWI_CONTEXT_MEMORY     phwi_context;
    ACIT_UEXACT32           desc_index, i;

    //
    // In step2, proceed to full initialization.
    //

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    // In the restricted mode there is no need to initialize any hardware rings.
    // Driver will use the mailbox mode to send IO. So return from here.
    if(pcontroller->state.api_type == ACIT_OP_MODE_RESTRICTED)
        return ACIT_STATUS_SUCCESS;

    phwi_controller->pmcc_wrb_base = (PMCC_WRB) pcontroller->
        mem_descriptors[HWI_MEM_MCC].virtual_address;

    phwi_controller->phwic = (PHWI_CONTEXT_MEMORY) pcontroller->
        mem_descriptors[HWI_MEM_ADDN_CONTEXT].virtual_address;

    phwi_controller->phwic->pbe_mcc_context = (PBE_MCC_WRB_CONTEXT) pcontroller->
        mem_descriptors[HWI_MEM_MCC_CONTEXT].virtual_address;

    desc_index = phwi_controller->frag_mem_index;
    OSM_ASSERT(pcontroller->acit_var_desc_index.num.frag <= HWI_MEM_NUM_FRAG);
    for(i=0; i<pcontroller->acit_var_desc_index.num.frag; i++, desc_index++)
    {
        phwi_controller->sge_base[i] = \
                pcontroller->mem_descriptors[desc_index].virtual_address;
    }

    desc_index = phwi_controller->wrb_mem_index;
    OSM_ASSERT(pcontroller->acit_var_desc_index.num.wrb <= HWI_MEM_NUM_WRB);
    for(i=0; i<pcontroller->acit_var_desc_index.num.wrb; i++, desc_index++)
    {
        phwi_controller->wrb_base[i] = \
                pcontroller->mem_descriptors[desc_index].virtual_address;
    }

    //
    // Initialize the context memory
    //

    phwi_context = phwi_controller->phwic;

    // The WRB queues are at the end of the context
    phwi_context->pbe_wrbq = (PBE_CXN_QUEUE_OBJECT)(phwi_context + 1);

    // Following the WRB queues, the EQ/CQ objects are located
    phwi_context->phwi_queue = (PHWI_INT_QUEUE_OBJECTS)&phwi_context->pbe_wrbq[pcontroller->params.num_cxn].be_wrbq;

    // Check context setup
    {
        ACIT_UEXACT32 ctx_len;
        PUEXACT32 cva, cva_end;
        cva = pcontroller->mem_descriptors[HWI_MEM_ADDN_CONTEXT].virtual_address;
        ctx_len = pcontroller->mem_descriptors[HWI_MEM_ADDN_CONTEXT].size_in_bytes;

        cva_end = (PUEXACT32)&phwi_context->phwi_queue[pcontroller->params.num_cq].be_eq;

        OSM_ASSERT(((ACIT_UEXACT32)(cva_end - cva)) <= ctx_len);
    }

    hwi_init_wrb_handle(pcontroller);

    if(hwi_init_beclib_ex(pcontroller, osm_handle, ACIT_DESC_TYPE_NON_INIT) != ACIT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("hwi_init_beclib_ex ACIT_DESC_TYPE_NON_INIT failed\n"));
        return ACIT_STATUS_ERROR;
    }

    // initialize the sgl handle now since we need hwi_init_beclib to get
    // the firmware config prior to initializing the sgl handle
    hwi_init_sgl_handle(pcontroller);

    // reserve a SGL for NOPOut.
    if(hwi_reserve_nop_sgl(pcontroller) != ACIT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("hwi_reserve_nop_sgl failed\n"));
        hwi_cleanup_beclib(pcontroller);
        return ACIT_STATUS_ERROR;
    }

    // Get PCIE Link info from CSR
    hwi_get_pci_link_info(pcontroller);

    // Initialize the translation rountines that handle the ASIC differences
    hwi_init_translation_routines(phwi_controller);

    return ACIT_STATUS_SUCCESS;
}

ACIT_U32
hwi_init_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle)
{
    ACIT_UEXACT32           status;

    // The legacy version calls the extended versions back to back
    
    status = hwi_init_controller_ex(pcontroller,
                                    osm_handle,
                                    ACIT_DESC_TYPE_INIT);

    if(!status)
    {
        status = hwi_init_controller_ex(pcontroller,
                                        osm_handle,
                                        ACIT_DESC_TYPE_NON_INIT);
    }

    return status;
}

ACIT_U32
hwi_init_controller_ex(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle,
    ACIT_UEXACT8            desc_type)
{
    ACIT_UEXACT32           status = ACIT_STATUS_ERROR;

    switch(desc_type)
    {
    case ACIT_DESC_TYPE_INIT:
        status = hwi_init_controller_step1(pcontroller, osm_handle);
        break;

    case ACIT_DESC_TYPE_NON_INIT:
        status = hwi_init_controller_step2(pcontroller, osm_handle);
        break;

    default:
        break;
    }
   
    return status;
}

ACIT_VOID
hwi_stop_controller(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc;
    MGMT_STATUS             mgmt_status;
    ACIT_UEXACT32           isr, pid, ulp;

    // initiate cleanup of EP firmware
    for(ulp = 0; ulp < BE_MAX_ULP; ulp++)
    {
        if(pcontroller->params.ulp_mask & (1 << ulp))
        {
            mgmt_status = mgmt_epfw_cleanup(pcontroller, ulp);
            OSM_ASSERT(mgmt_status == MGMT_STATUS_SUCCESS);
        }
    }

    // destroy all rings
    hwi_cleanup_beclib(pcontroller);

    // purge all EQ entries that may have been left out
    // This is to workaround a problem we've seen occasionally where
    // we exit from acit_stop_controller and we get an interrupt
    // with EQ entry bit set
    for(pid = 0; pid < pcontroller->params.num_cq; pid++)
    {
        hwi_purge_eq(pcontroller, pid);
    }

    // Read the ISR register to clear any pending interrupt bit
    isr = CsrRead(pcontroller, pcontroller->fw_config.isr_offset);
    
    // destroy SA device object
    phwc = GET_HWI_CONTROLLER_WS(pcontroller);
    sa_dev_destroy (&phwc->phwic->sa_dev);
}

ACIT_UEXACT32
hwi_register_arm_async_events(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc;
    PHWI_CONTEXT_MEMORY     phwi_context;
    BESTATUS                bestatus = 0;

    phwc = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwc);
    phwi_context = phwc->phwic;

    bestatus = be_mcc_add_async_event_callback(
                                &phwi_context->be_mcc,
                                (MCC_ASYNC_EVENT_CALLBACK)mgmt_async_event_handler,
                                pcontroller);

    if (bestatus != BE_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Async Event registration failed!!\n"));
        return ACIT_STATUS_ERROR;
    }

    phwi_context->initialized.async_regn = 1; 
    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_check_and_log_ue(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PCICFG_UE_STATUS_LOW_CSR ue_lo;
    PCICFG_UE_STATUS_HI_CSR  ue_hi;
    ACIT_UEXACT32 ue_lo_mask, ue_hi_mask; 
    ACIT_UEXACT32 semaphore_offset = 0, fw_line = ACIT_U32_MINUS_ONE;
    ACIT_UEXACT32 ext_error[3];

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Run UE Detection\n"));

    ue_lo.dw = ACIT_PCICFG0_READ(pcontroller, ue_status_low);
    ue_hi.dw = ACIT_PCICFG0_READ(pcontroller, ue_status_hi);

    // Bug 109828 : Use the cached value of mask registers to detect a UE
    ue_lo.dw = (ue_lo.dw & (~pcontroller->ue_detection.cached_ue0_mask));
    ue_hi.dw = (ue_hi.dw & (~pcontroller->ue_detection.cached_ue1_mask));

    if(ue_lo.dw || ue_hi.dw)
    {
        // We are in UE!! Determine the offset

        if(ue_lo.rxulp0_ue_status)
            semaphore_offset = RXULP0_SEMAPHORE_OFFSET;
        else if(ue_lo.rxulp1_ue_status)
            semaphore_offset = RXULP1_SEMAPHORE_OFFSET;
        else if(ue_lo.rxulp2_ue_status)
            semaphore_offset = RXULP2_SEMAPHORE_OFFSET;
        else if(ue_lo.tpost_ue_status)
            semaphore_offset = TPOST_SEMAPHORE_OFFSET;
        else if(ue_lo.tpre_ue_status)
            semaphore_offset = TPRE_SEMAPHORE_OFFSET;
        else if(ue_lo.txulp0_ue_status)
            semaphore_offset = TXULP0_SEMAPHORE_OFFSET;
        else if(ue_lo.txulp1_ue_status)
            semaphore_offset = TXULP1_SEMAPHORE_OFFSET;
        else if(ue_lo.txulp2_ue_status)
            semaphore_offset = TXULP2_SEMAPHORE_OFFSET;
        else if(ue_lo.mpu_ue_status)
            semaphore_offset = MPU_SEMAPHORE_OFFSET;

        fw_line = CsrRead(pcontroller, semaphore_offset);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_FATAL,(("UE Detected - UE Low 0x%x UE Hi 0x%x Line# 0x%x\n"),
                            ue_lo.dw, ue_hi.dw, fw_line));

        ext_error[0] = ue_lo.dw;
        ext_error[1] = ue_hi.dw;
        ext_error[2] = fw_line;

        ap_log_osm_event(pcontroller,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         3,
                         ACIT_EC_HWI_UE,
                         0, ACIT_NULL,
                         &ext_error[0]);

        pcontroller->eh_stats.ue_low = ue_lo.dw;
        pcontroller->eh_stats.ue_hi = ue_hi.dw;
        pcontroller->eh_stats.ue_fw_line = fw_line;
        pcontroller->eh_stats.ue_detected = 1;

        OSM_ASSERT(0);
    }

    if(pcontroller->eh_stats.ue_detected)
        return ACIT_STATUS_ERROR;
    else
        return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_check_health(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           log_error)
{
    ACIT_UEXACT32           run_detection = 1;
    ACIT_UEXACT32           status;

    //
    // Check if there were any IO or IOCTL completions from the last time we
    // ran. If so we can defer the polling.. We also do not need to run
    // detection again if ue_detected has been set.
    // Note! For UE Recovery, it is assumed the flag is turned off when the
    // recovery is complete
    //
    if((pcontroller->stats.ios_completed > 
        pcontroller->ue_detection.last_io_completed) ||
       (pcontroller->stats.ioctls_completed > 
        pcontroller->ue_detection.last_ioctl_completed))
    {
        run_detection = 0;
    }

    if(ACIT_IS_UE_RECOVERY_ENABLED(pcontroller))
    {
        if(pcontroller->ue_detection.ue_recovery_started)
        {
            // UE recovery has been initiated. Just poll for completion

            run_detection = 0;

            status = hwi_poll_for_ue_recovery(pcontroller);

            if(status != ACIT_STATUS_PENDING)
            {
                ap_complete_ue_recovery(pcontroller, status);
            }
        }
        else
        {
            // UE recovery not started. Run the detection. aka NO-OP
        }
    }
    else
    {
        // In UE tolerance. Skip the UE detection if we are already in UE
        if(pcontroller->eh_stats.ue_detected)
        {
            OSM_ASSERT(pcontroller->counters.queued_tmfs == 0);
            OSM_ASSERT(pcontroller->counters.pending_tmfs == 0);
            OSM_ASSERT(pcontroller->counters.queued_ios == 0);
            OSM_ASSERT(pcontroller->counters.pending_ios == 0);

            run_detection = 0;
        }
    }

    // Update the housekeeping stats
    pcontroller->ue_detection.last_io_completed = pcontroller->stats.ios_completed;
    pcontroller->ue_detection.last_ioctl_completed = pcontroller->stats.ioctls_completed;

    if(run_detection)
    {
        if(ACIT_STATUS_SUCCESS != hwi_check_and_log_ue(pcontroller))
        {
            // Handle UE occurrance
            ap_process_ue_occurrence(pcontroller);
        }
    }
}

//
// This routine dynamically configures the EQ timer based on
// the interrupt rate. This is run for every timer tick in ACIT
//

ACIT_VOID
hwi_check_interrupts(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           pid)
{
    ACIT_UEXACT32           reconfigure = 0;
    ACIT_SEXACT32           new_eq_delay = 0;
    ACIT_UEXACT32           io_load = 0; 
    ACIT_UEXACT32           load_factor = 0, target_load = 0;

    pcontroller->partition[pid].dic.interrupts_per_sec = \
                (pcontroller->partition[pid].dic.interrupts - 
                 pcontroller->partition[pid].dic.last_interrupt_count) *
                TICKS_PER_SEC(pcontroller);

    if(pcontroller->partition[pid].dic.interrupts_per_sec > 
       pcontroller->partition[pid].dic.max_interrupts_per_sec)
    {
        pcontroller->partition[pid].dic.max_interrupts_per_sec = 
                        pcontroller->partition[pid].dic.interrupts_per_sec;
    }

    // extrapolate to ios_per second
    pcontroller->partition[pid].dic.ios_per_sec *= TICKS_PER_SEC(pcontroller);

    if(pcontroller->partition[pid].dic.ios_per_sec > 
       pcontroller->partition[pid].dic.max_ios_per_sec)
    {
        pcontroller->partition[pid].dic.max_ios_per_sec = 
            pcontroller->partition[pid].dic.ios_per_sec;
    }

   {
        ACIT_UEXACT32 eq_run;

        eq_run = pcontroller->partition[pid].dic.eq_timer / pcontroller->params.dic.timer_increments;

        if(eq_run <= 31)
        {
            pcontroller->partition[pid].dic.eq_runs[eq_run]++;
        }
    }

    ap_determine_load(pcontroller, &io_load, &load_factor, &target_load, pid);

    pcontroller->partition[pid].dic.current_target_load = target_load;
    pcontroller->partition[pid].dic.current_load_factor = load_factor;

    if(!io_load)
    {
        // if there is no load, then the interrupt rate and timers must go back to default
        if((pcontroller->partition[pid].dic.interrupt_rate != pcontroller->params.dic.max_interrupt_rate) ||
           (pcontroller->partition[pid].dic.eq_timer != pcontroller->params.dic.min_timer_val))
        {
            pcontroller->partition[pid].dic.interrupt_rate = pcontroller->params.dic.max_interrupt_rate;
            new_eq_delay = pcontroller->params.dic.min_timer_val;
            reconfigure = 1;
        }
    }
    else
    {
        pcontroller->partition[pid].dic.interrupt_rate = pcontroller->params.dic.max_interrupt_rate -
                                            target_load * pcontroller->params.dic.interrupt_dec_steps;

        if(pcontroller->partition[pid].dic.interrupt_rate < pcontroller->params.dic.min_interrupt_rate)
        {
            pcontroller->partition[pid].dic.interrupt_rate = pcontroller->params.dic.min_interrupt_rate;
        }

        if((ACIT_SEXACT32)pcontroller->partition[pid].dic.interrupts_per_sec > 
            (pcontroller->partition[pid].dic.interrupt_rate + (ACIT_SEXACT32)pcontroller->params.dic.interrupt_resolution))
        {
            // is the system loaded at the same time with lots of IOs?
            // In this case modify eq timer thresholds at a faster rate and also
            // modify the expected interrupt rate

            new_eq_delay = pcontroller->partition[pid].dic.eq_timer + 
                            pcontroller->params.dic.timer_increments * load_factor;

            if(new_eq_delay > (ACIT_SEXACT32)pcontroller->params.dic.max_timer_val)
                new_eq_delay = (ACIT_SEXACT32)pcontroller->params.dic.max_timer_val;

            reconfigure = 1;
        }
        else if((ACIT_SEXACT32)pcontroller->partition[pid].dic.interrupts_per_sec < 
                (pcontroller->partition[pid].dic.interrupt_rate - (ACIT_SEXACT32)pcontroller->params.dic.interrupt_resolution))
        {
            new_eq_delay = pcontroller->partition[pid].dic.eq_timer - pcontroller->params.dic.timer_increments;

            if(new_eq_delay < (ACIT_SEXACT32)pcontroller->params.dic.min_timer_val)
                new_eq_delay = (ACIT_SEXACT32)pcontroller->params.dic.min_timer_val;

            reconfigure = 1;
        }
    }

    if(reconfigure && 
       (new_eq_delay != (ACIT_SEXACT32)(pcontroller->partition[pid].dic.eq_timer)))
    {
        PBE_FUNCTION_OBJECT pfo;
        ACIT_UEXACT32 delay_multiplier;
        OSM_ASSERT(new_eq_delay);

        pfo = hwi_get_be_pfo(pcontroller);
        delay_multiplier = be_eq_delay_to_multiplier(pfo, new_eq_delay);
        
        mgmt_modify_eq_delay(pcontroller, 
                             pid,
                             HWI_GET_EQ_ID(pcontroller, pid),
                             delay_multiplier);

        pcontroller->partition[pid].dic.eq_timer = new_eq_delay;            
    }

    // make this the new interrupt rate
    pcontroller->partition[pid].dic.last_interrupt_count = pcontroller->partition[pid].dic.interrupts;

    pcontroller->partition[pid].dic.last_ios_per_sec = pcontroller->partition[pid].dic.ios_per_sec;
    pcontroller->partition[pid].dic.ios_per_sec = 0;

}   

#if 0

ACIT_VOID
hwi_check_interrupts(
    PACIT_CONTROLLER_WS     pcontroller)
{
    ACIT_UEXACT32           reconfigure = 0;
    ACIT_SEXACT32           new_eq_delay = 0;
    ACIT_UEXACT32           io_load = 0; 
    ACIT_UEXACT32           load_factor = 0, target_load = 0;

    pcontroller->dic.interrupts_per_sec = \
                (pcontroller->dic.interrupts - pcontroller->dic.last_interrupt_count) *
                TICKS_PER_SEC(pcontroller);

    if(pcontroller->dic.interrupts_per_sec > pcontroller->dic.max_interrupts_per_sec)
    {
        pcontroller->dic.max_interrupts_per_sec = pcontroller->dic.interrupts_per_sec;
    }

    // extrapolate to ios_per second
    pcontroller->dic.ios_per_sec *= TICKS_PER_SEC(pcontroller);

    if(pcontroller->dic.ios_per_sec > pcontroller->dic.max_ios_per_sec)
    {
        pcontroller->dic.max_ios_per_sec = pcontroller->dic.ios_per_sec;
    }

    {
        ACIT_UEXACT32 eq_run;

        eq_run = pcontroller->dic.eq_timer / pcontroller->params.dic.timer_increments;

        if(eq_run <= 31)
        {
            pcontroller->dic.eq_runs[eq_run]++;
        }
    }

    ap_determine_load(pcontroller, &io_load, &load_factor, &target_load);

    pcontroller->dic.current_target_load = target_load;
    pcontroller->dic.current_load_factor = load_factor;

    if(!io_load)
    {
        // default to max interrupt rate and min eq timer when there's no load
        if((pcontroller->dic.interrupt_rate != pcontroller->params.dic.max_interrupt_rate) ||
            (pcontroller->dic.eq_timer != pcontroller->params.dic.min_timer_val))
        {
            pcontroller->dic.interrupt_rate = pcontroller->params.dic.max_interrupt_rate;
            new_eq_delay = pcontroller->params.dic.min_timer_val;
            reconfigure = 1;
        }
    }
    else
    {
        pcontroller->dic.interrupt_rate = pcontroller->params.dic.max_interrupt_rate -
                                            target_load * pcontroller->params.dic.interrupt_dec_steps;

        if(pcontroller->dic.interrupt_rate < pcontroller->params.dic.min_interrupt_rate)
        {
            pcontroller->dic.interrupt_rate = pcontroller->params.dic.min_interrupt_rate;
        }

        if((ACIT_SEXACT32)pcontroller->dic.interrupts_per_sec > 
           (pcontroller->dic.interrupt_rate + pcontroller->params.dic.interrupt_resolution))
        {
            // is the system loaded at the same time with lots of IOs?
            // In this case modify eq timer thresholds at a faster rate and also
            // modify the expected interrupt rate

            new_eq_delay = pcontroller->dic.eq_timer + pcontroller->params.dic.timer_increments * load_factor;

            if(new_eq_delay > (ACIT_SEXACT32)pcontroller->params.dic.max_timer_val)
                new_eq_delay = (ACIT_SEXACT32)pcontroller->params.dic.max_timer_val;

            reconfigure = 1;
        }
        else if((ACIT_SEXACT32)pcontroller->dic.interrupts_per_sec <
                (pcontroller->dic.interrupt_rate - pcontroller->params.dic.interrupt_resolution))
        {
            new_eq_delay = pcontroller->dic.eq_timer - pcontroller->params.dic.timer_increments;

            if(new_eq_delay < (ACIT_SEXACT32)pcontroller->params.dic.min_timer_val)
                new_eq_delay = (ACIT_SEXACT32)pcontroller->params.dic.min_timer_val;

            reconfigure = 1;
        }
    }

    if(reconfigure && (new_eq_delay != pcontroller->dic.eq_timer))
    {
        PBE_FUNCTION_OBJECT pfo;
        ACIT_UEXACT32 delay_multiplier;
        OSM_ASSERT(new_eq_delay);

        pfo = hwi_get_be_pfo(pcontroller);
        delay_multiplier = be_eq_delay_to_multiplier(pfo, new_eq_delay);
        
        mgmt_modify_eq_delay(pcontroller,
                             0,
                             pcontroller->phwi_ws->eq.id,
                             delay_multiplier);

        pcontroller->dic.eq_timer = new_eq_delay;            
    }

    // make this the new interrupt rate
    pcontroller->dic.last_interrupt_count = pcontroller->dic.interrupts;

    pcontroller->dic.last_ios_per_sec = pcontroller->dic.ios_per_sec;
    pcontroller->dic.ios_per_sec = 0;

}   

#endif

ACIT_VOID
hwi_build_sa_sgl(
    IN  PACIT_CONTROLLER_WS pcontroller,
    IN  ACIT_UEXACT32       mem_descriptor_index,
    OUT PSA_SGL             sgl)
{
    PHWI_CONTROLLER_WS  phwi_controller;

    ASSERT (sgl);
    ASSERT (pcontroller);
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwi_controller);

    // Destroy any existing SGL.
    if (sa_sgl_get_base_va (sgl)) {
        sa_sgl_destroy_contiguous (sgl);
    }

    sa_sgl_create_contiguous (
        pcontroller->mem_descriptors[mem_descriptor_index].virtual_address,
        pcontroller->mem_descriptors[mem_descriptor_index].bus_address.u.a64.address,
        pcontroller->mem_descriptors[mem_descriptor_index].size_in_bytes,
        sgl);
}

ACIT_VOID
hwi_build_sa_sgl_by_offset (
    IN  PACIT_CONTROLLER_WS pcontroller,
    IN  ACIT_UEXACT32       mem_descriptor_index,
    IN  ACIT_UEXACT32       offset,
    IN  ACIT_UEXACT32       size,
    OUT PSA_SGL             sgl)
{
    PHWI_CONTROLLER_WS  phwi_controller;

    ASSERT (sgl);
    ASSERT (pcontroller);
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwi_controller);

    // Destroy any existing SGL.
    if (sa_sgl_get_base_va (sgl)) {
        sa_sgl_destroy_contiguous (sgl);
    }

    sa_sgl_create_contiguous (
        (PACIT_UEXACT8)pcontroller->mem_descriptors[mem_descriptor_index].virtual_address + offset,
        pcontroller->mem_descriptors[mem_descriptor_index].bus_address.u.a64.address + offset,
        size,
        sgl);
}

ACIT_VOID
hwi_cleanup_beclib(
    PACIT_CONTROLLER_WS     pcontroller)    
{
    PHWI_CONTROLLER_WS      phwc = NULL;
    PHWI_CONTEXT_MEMORY     phwi_context;
    BESTATUS bestatus = 0;

    ASSERT (pcontroller);
    phwc = GET_HWI_CONTROLLER_WS (pcontroller);
    ASSERT (phwc);
    phwi_context = phwc->phwic;
    OSM_ASSERT(phwi_context);

    bestatus = be_function_cleanup(&phwi_context->be_function);
    OSM_ASSERT(bestatus == BE_SUCCESS);

    return;
}

ACIT_UEXACT32
hwi_init_beclib_ex(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle,
    ACIT_UEXACT8            desc_type)
{
    ACIT_UEXACT32           status = ACIT_STATUS_ERROR;

    switch(desc_type)
    {
    case ACIT_DESC_TYPE_INIT:
        status = hwi_init_beclib_step1(pcontroller, osm_handle);
        break;

    case ACIT_DESC_TYPE_NON_INIT:
        status = hwi_init_beclib_step2(pcontroller, osm_handle);
        break;

    default:
        break;
    }

    return status;
}

ACIT_UEXACT32
hwi_init_beclib_step1(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle)
{
    PACIT_CONTROLLER_WS     pc;
    PHWI_CONTROLLER_WS      phwi;
    PHWI_CONTEXT_MEMORY     pcontext;
    PBE_FUNCTION_OBJECT     pfo;
    ACIT_UEXACT32           status;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwi);

    pcontext = phwi->phwic;
    OSM_ASSERT(pcontext);

    // abbreviations
    pc = pcontroller;
    pfo = &pcontext->be_function;

    if(pc->ue_detection.reinit_after_ue == 1)
    {
        // If this routine is called because of reinit after UE, go directly
        // to step2
        return ACIT_STATUS_SUCCESS;
    }

    status = hwi_init_mem_bars(pc, phwi, pcontext, osm_handle);
    if(status != 0)
    {
        goto error;
    }

check_ue_state:
    status = hwi_check_ue_at_init(pc, phwi, pcontext);
    if(status != 0)
    {
        goto error;
    }

init_start:

     // BEClib library initialization.
    status = hwi_init_belib_library(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    // Some initialization sequence needs a hardware specific
    // reset.
    status = hwi_check_and_issue_reset(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    if(pc->state.soft_reset)
    {
        // If a soft reset has been initiated, restart the initialization of
        // library since the soft reset destroys the objects created by the
        // library
        goto init_start;
    }

    if(pc->eh_stats.ue_detected)
    {
        // If there was a UE, go back and handle it. Depending on
        // the UE recovery settings, it will be handled appropriately
        goto check_ue_state;
    }

#ifdef BE_EMULATION
    // initialize for emulation
    if(emul_hwi_init(pc) != ACIT_STATUS_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("Initialization failed for emulation."));
        goto error;
    }
#endif

    // Read important chip related attributes. Note! Only after this API will
    // the ASIC generation be populated in mgmt_attribute structure.
    status = hwi_init_mgmt_attrib(pc);
    if(status != 0)
    {
        goto error;
    }

error:

    if (status != 0) {
        TRACE (DL_ERR, "beclib init failed");

        if(pcontext->initialized.chip_function)
        {
            hwi_cleanup_beclib (pc);
        }

        ap_log_osm_event(pc,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         1,
                         ACIT_EC_HWI_INIT_FAILED,
                         0, ACIT_NULL,
                         &status);

        return ACIT_STATUS_ERROR;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
hwi_init_beclib_step2(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle)
{
    PACIT_CONTROLLER_WS     pc;
    PHWI_CONTROLLER_WS      phwi;
    PHWI_CONTEXT_MEMORY     pcontext;
    PBE_FUNCTION_OBJECT     pfo;
    ACIT_UEXACT32           status;

    phwi = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwi);

    pcontext = phwi->phwic;
    OSM_ASSERT(pcontext);

    pc = pcontroller;

    // In the restricted mode there is no need to initialize any hardware rings.
    // Driver will use the mailbox mode to send IO. So return from here.
    if(pc->state.api_type == ACIT_OP_MODE_RESTRICTED)
        return ACIT_STATUS_SUCCESS;

    pfo = &pcontext->be_function;

    // Create software ring objects for managing EQ, CQ, RQs
    hwi_init_sa_rings(pc, phwi);

    // Create hardware EQ
    status = hwi_init_eq(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    // Create MCC mode rings
    status = hwi_init_mcc_rings(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    // Create hardware CQs
    status = hwi_init_cq(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    // Create hardware RQs
    status = hwi_init_rq(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    // Post template headers
    status = hwi_init_template_headers(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    // Initialize SGL
    status = hwi_init_sgl_pages(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    // Initialize WRB rings
    status = hwi_init_wrb_rings(pc, phwi, pcontext, pfo);
    if(status != 0)
    {
        goto error;
    }

    // Vase specific initialization
    hwi_init_vase_memory(pc, phwi);

error:

    if (status != 0) {
        TRACE (DL_ERR, "beclib init failed");

        if(pcontext->initialized.chip_function)
        {
            hwi_cleanup_beclib (pc);
        }

        ap_log_osm_event(pc,
                         UNI_CFG_LOG_LEVEL_CRITICAL,
                         UNI_SUBMODULE_ACIT_HWI,
                         UNI_PHASE_DONT_CARE,
                         1,
                         ACIT_EC_HWI_INIT_FAILED,
                         0, ACIT_NULL,
                         &status);

        return ACIT_STATUS_ERROR;
    }

    return ACIT_STATUS_SUCCESS;

}
ACIT_U8
hwi_disable_irq(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PBE_FUNCTION_OBJECT     pfo;

    if(pcontroller->state.api_type != ACIT_OP_MODE_RESTRICTED)
    {
        pfo = hwi_get_be_pfo(pcontroller);
        OSM_ASSERT(pfo);
        be_function_disable_interrupts(pfo);
    }
    else
    {
        // not handled currently. This will be enabled once we determine
        // how the crashdump mode will be handled
        OSM_ASSERT(0);
    }

    pcontroller->state.interrupt_enabled = ACIT_FALSE;
    return ACIT_TRUE;
}

ACIT_U8
hwi_enable_irq(
    PACIT_CONTROLLER_WS     pcontroller)

{
    PCICFG_MEMBAR_CTRL_INT_CTRL_CSR ctrl;   
    CEV_ISR0_CSR       isr;
    ACIT_UEXACT32       pid = 0;

    ctrl.dw = ACIT_PCICFG0_READ (pcontroller, membar_ctrl_intr_ctrl);

    isr.dw = CsrRead(pcontroller, pcontroller->fw_config.isr_offset);

    if (!ctrl.hostintr) 
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("HostIntr not set, setting now..\n"));

        ctrl.hostintr = 1;        
        ACIT_PCICFG0_WRITE(pcontroller, membar_ctrl_intr_ctrl, ctrl.dw);        
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("HostIntr set, only consuming pending eq..\n"));
    }

    for(pid = 0; pid < pcontroller->params.num_cq; pid++)
    {
        //
        // Bug 9732: If the EQ is not processed and rearmed, when resuming from
        // local boot hibernation, we stop receiving interrupts. Its also important
        // to have the correct value for num_processed. The beclib function that
        // enables interrupt always writes a value of 0 for num_processed. Therefore
        // we write our own enable interrupt routine
        //
        //hwi_process_eq(pcontroller, pid);

        // Bug 22807 : Cleanup MCC related EQ entry. If we process ISCSI_CQ entries
        // also, then the interrupts stay disabled forever.
        hwi_process_mcc_eq_from_init(pcontroller, pid);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Rearming EQ - id 0x%x num_processed %d\n",
            HWI_GET_EQ_ID(pcontroller, pid), pcontroller->partition[pid].dic.num_eq_processed));

        // Rearm the EQ since we may have missed the last interrupt while
        // the host interrupts were disabled.
        hwi_ring_eq_db (pcontroller, 
                        HWI_GET_EQ_ID(pcontroller, pid),
                        0, // clear interrupt
                        pcontroller->partition[pid].dic.num_eq_processed, 
                        1, 
                        1);

        pcontroller->partition[pid].dic.num_eq_processed = 0;
    }

    pcontroller->state.interrupt_enabled = ACIT_TRUE;
    return ACIT_TRUE;
}

ACIT_VOID
hwi_process_mcc_cq(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    be_mcc_process_cq(&phwi_controller->phwic->be_mcc, ACIT_FALSE);   // no-rearm
}

ACIT_VOID
hwi_purge_eq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    SA_RING *ring;
    EQ_ENTRY *eqe; 

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &pcontroller->partition[message_id].eq;
    eqe = (EQ_ENTRY*) sa_ring_current (ring); 

    //
    // Zero out all valid entries
    //

    while (eqe->Valid) {

        //
        // Mark this EQ_ENTRY as invalid
        //
        eqe->Valid = 0;

        //
        // Circularly increment the consumer index, and move to next EQ_ENTRY
        // to process.
        //
        eqe = (EQ_ENTRY*) sa_ring_next (ring);
    }
}


ACIT_VOID
hwi_process_eq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    SA_RING *ring;
    EQ_ENTRY *eqe; 

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_eq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &pcontroller->partition[message_id].eq;

    eqe = (EQ_ENTRY*) sa_ring_current (ring); 

    ACIT_UPDATE_INT_COUNT(pcontroller, message_id);

    //
    // Process the Event Queue
    //

    DBG_LOG_FUNC_EXEC(&pcontroller->vase_master_block,
                        FUNC_hwi_process_eq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);

    while (eqe->Valid) {
        switch (eqe->MajorCode) {
            case EQ_MAJOR_CODE_COMPLETION:
                //
                // Determine which CQ to process.
                //
                if (eqe->ResourceID == HWI_GET_CQ_ID(pcontroller, message_id)) {
                    
                    HWI_PROCESS_CQ (pcontroller, message_id);
                    ACIT_UPDATE_EQ_MAJ_COMP_INT_COUNT(pcontroller, message_id);

                } 
                else
                {
                    be_mcc_process_cq(&phwi_controller->phwic->be_mcc,
                                      ACIT_TRUE);   // rearm
                }
                break;

            default:
                OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("Unexpected EQ_ENTRY MajorCode: 0x%x\n", eqe->MajorCode));
                pcontroller->partition[message_id].interrupt_stats.unknown_eqes++;
                OSM_ASSERT (0);
        }

        //
        // Mark this EQ_ENTRY as invalid
        //
        eqe->Valid = 0;

        //
        // Circularly increment the consumer index, and move to next EQ_ENTRY
        // to process.
        //
        eqe = (EQ_ENTRY*) sa_ring_next (ring);

        pcontroller->partition[message_id].dic.num_eq_processed++;
    }

    // assert on no EQ entries found
    if(pcontroller->state.interrupt_enabled)
    {
        if(pcontroller->partition[message_id].dic.num_eq_processed == 0)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("No valid EQ entries on message_id %d\n",
                                            message_id));
            // OSM_ASSERT(0);
        }
    }

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_eq,
                        VASE_MODULE_ACIT,
                        0,
                        0,
                        0);
}

ACIT_VOID
hwi_process_mcc_eq_from_init (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    SA_RING *ring;
    EQ_ENTRY *eqe; 

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &pcontroller->partition[message_id].eq;

    eqe = (EQ_ENTRY*) sa_ring_current (ring); 

    //
    // Process only the first EQE. It is ASSUMED that the first entry is
    // the one arising out of completing initialization IOCTLs via polling
    // the MCC CQ directly.
    //

    if (eqe->Valid) {
        switch (eqe->MajorCode) {
            case EQ_MAJOR_CODE_COMPLETION:
                //
                // Determine which CQ to process.
                //
                if (eqe->ResourceID == HWI_GET_CQ_ID(pcontroller, message_id)) {
                    
                    OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("Unexpected ResourceID: 0x%x\n", eqe->ResourceID));
                    ACIT_UPDATE_EQ_MAJ_COMP_INT_COUNT(pcontroller, message_id);
                    OSM_ASSERT(0);

                    // Jump to exit. We will not update the consumer index and
                    // nor will we increment num_eq_processed
                    goto end_eq_process;
                } 
                else
                {
                    be_mcc_process_cq(&phwi_controller->phwic->be_mcc,
                                      ACIT_TRUE);   // rearm
                }
                break;

            default:
                OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("Unexpected EQ_ENTRY MajorCode: 0x%x\n", eqe->MajorCode));
                pcontroller->partition[message_id].interrupt_stats.unknown_eqes++;
                OSM_ASSERT (0);
        }

        //
        // Mark this EQ_ENTRY as invalid
        //
        eqe->Valid = 0;

        //
        // Circularly increment the consumer index, and move to next EQ_ENTRY
        // to process.
        //
        eqe = (EQ_ENTRY*) sa_ring_next (ring);

        pcontroller->partition[message_id].dic.num_eq_processed++;
    }

end_eq_process:

    return;
}

ACIT_U32
hwi_process_irq(
    PACIT_CONTROLLER_WS     pcontroller)
{
    CEV_ISR0_CSR       isr;

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);

    //
    // Read interrupt status register to determine source of interrupt.
    // This is a read clear register so we must process all posted interrupts
    // before exiting the interrupt handler.
    // 
    isr.dw = CsrRead(pcontroller, pcontroller->fw_config.isr_offset);

    DBG_LOG_FUNC_EXEC(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        isr.dw,
                        0);

    if (!isr.dw)
    {
        if(pcontroller->state.interrupt_enabled)
        {
            //
            // This is not our interrupt.
            // 
            pcontroller->partition[0].interrupt_stats.unknown_ints++;
            DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                                FUNC_hwi_process_irq,
                                VASE_MODULE_ACIT,
                                pcontroller->counters.pending_ios,
                                0,
                                0);

            DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                        InvalidInterrupts);
        }

        return ACIT_STATUS_UNKNOWN_INTERRUPT;      
    }


    hwi_process_eq (pcontroller, 0);

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);
    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                ValidInterrupts);

    return ACIT_STATUS_SUCCESS;
}

ACIT_U32
hwi_process_msi_irq(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id)
{

    DBG_LOG_FUNC_ENTER(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);

    hwi_process_eq (pcontroller, message_id);

    DBG_LOG_FUNC_EXIT(&pcontroller->vase_master_block,
                        FUNC_hwi_process_irq,
                        VASE_MODULE_ACIT,
                        pcontroller->counters.pending_ios,
                        0,
                        0);
    DBG_LOG_STATISTICS_INCREMENT(&pcontroller->vase_master_block,
                                ValidInterrupts);

    return ACIT_STATUS_SUCCESS;
}


ACIT_STATUS
hwi_alloc_eh_resources(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri,
    PWRB_HANDLE             *pwrb_handle,
    PSGL_HANDLE             *psgl_handle)
{
    PWRB_HANDLE             pwrb_h = ACIT_NULL;
    PSGL_HANDLE             psgl_h = ACIT_NULL;

    OSM_ASSERT(phwi_controller);
    OSM_ASSERT(pwrb_handle);
    OSM_ASSERT(psgl_handle);

    pwrb_h = hwi_alloc_wrb_handle(phwi_controller, cri);

    if(pwrb_h)
    {
        psgl_h = hwi_alloc_eh_sgl_handle(phwi_controller);
        if(psgl_h) 
        {
            *pwrb_handle = pwrb_h;
            *psgl_handle = psgl_h;
            return ACIT_STATUS_SUCCESS;
        }
        hwi_free_wrb_handle(phwi_controller, cri, pwrb_h);
    }

    OSM_ASSERT(0);
    return ACIT_STATUS_INSUFFICIENT_RESOURCE;
}

ACIT_VOID
hwi_free_eh_resources(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri,
    PWRB_HANDLE             pwrb_handle,
    PSGL_HANDLE             psgl_handle)
{
    hwi_free_wrb_handle(phwi_controller, 
                        cri, 
                        pwrb_handle);

    pwrb_handle->type = HWH_TYPE_INVALID;
    pwrb_handle->pio_handle = ACIT_NULL;

    hwi_free_eh_sgl_handle(phwi_controller, psgl_handle);
}

#ifdef ACIT_CFG_CHAINED_SGL
ACIT_VOID
hwi_free_chained_sgls(
    PHWI_CONTROLLER_WS  phwi_controller,
    PACIT_IO_HANDLE     pio_handle)
{
    PSGL_HANDLE         psgl_handle;
    PACIT_DC_LIST       psgh_link, tmp;

    for(psgh_link = pio_handle->psglh_list.next, tmp = psgh_link->next;
        psgh_link != &pio_handle->psglh_list;
        psgh_link = tmp, tmp = psgh_link->next) {

        psgl_handle = STRUCT_BASE(SGL_HANDLE,
                             link,
                             psgh_link);

        OSM_ASSERT(psgl_handle);

        LIST_INIT_ELEMENT(&psgl_handle->link);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("Freeing SGL %d - IOH 0x%p\n",
            psgl_handle->sgl_index, pio_handle));

        hwi_free_io_sgl_handle(phwi_controller, psgl_handle);
    }
}

ACIT_VOID
hwi_alloc_chained_sgls(
    PACIT_CONTROLLER_WS pcontroller, 
    PACIT_IO_HANDLE     pio_handle, 
    PSGL_HANDLE         * psgl_handle)
{
    PACIT_IO_REQUEST    pio_request;
    PACIT_SG_LIST       pacit_sgl;
    ACIT_UEXACT32       num_sge, num_sgls = 0;
    PSGL_HANDLE         plocal;
    ACIT_UEXACT32       i;

    // Allocate SGL handle based on the IO SGL list AND the configuration parameter
    *psgl_handle = ACIT_NULL;

    pio_request = pio_handle->pio->pio_request;
    pacit_sgl = pio_request->psg_list;

    if(pacit_sgl)
    {
        // We need at least as many SGEs as defined by ACIT
        num_sge = pacit_sgl->number_elements;
        num_sge +=2; // Need 2 more. One for BHS and other for SKIP SGE

        // In addition, if some ACIT_SGE is > 64K, then we have to split 'm up
        // into multiple sges. So count that here
        if(pio_request->data_transfer_length > BE_MAX_BYTES_PER_FRAG)
        {
            PACIT_SG_ELEMENT acit_sge;
            acit_sge = pacit_sgl->sglist;

            // !!!Warn!!!
            // The inefficient loop. We could move the allocation and fillup together.
            // But for now this is only for test, so live with this.
            for(i = 0; i < pacit_sgl->number_elements; i++)
            { 
                if(acit_sge->length > BE_MAX_BYTES_PER_FRAG)
                {
                    num_sge += (acit_sge->length / BE_MAX_BYTES_PER_FRAG);
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("IOH 0x%p has an SGE > 64!\n", pio_handle));
                }
                acit_sge++;
            }
        }

        num_sgls = num_sge/(pcontroller->params.num_sge-1);
        if(num_sge % (pcontroller->params.num_sge-1))
            num_sgls++;

        // Since we can use an incremental offset of frag_num for testing,
        // allocate one more SGL
        switch(pcontroller->dma_test.test_case)
        {
            case DMA_TEST_MULTIPLE_CHAINS_PER_SGL:
            {
                ACIT_UEXACT32 chains;
                // For every 3 SGEs, there needs to be a chain element
                chains = (num_sge / 3) + 1;
                // There can be max of 8 chains per SGL of 32 SGEs
                num_sgls = (chains / 8) + 1;
                break;
            }

            //case DMA_TEST_INC_FRAG_NUM:
            //case DMA_TEST_SEQINC_FRAG_NUM:
            default:
                // Allocate 2 more than what was previously allocated
                // Depending on the frag_num offset, one SGL may contain
                // not more than 2 elements. So allocate 2 extra SGLs
                num_sgls+=2;
                break;
        }
    }
    else
    {
        // If there is no SGL provided by the OS, then set the number of SGLs to
        // 1. We need atleast 1 ICD to issue an IO
        num_sgls = 1;
    }

    LIST_INIT_HEADER(&pio_handle->psglh_list);

    for(i = 0; i < num_sgls; i++)
    {
        plocal = hwi_alloc_io_sgl_handle(pcontroller->phwi_ws);

        if(plocal)
        {
            if(*psgl_handle == ACIT_NULL)
            {
                // Initialize the first element that we want to return
                *psgl_handle = plocal;
            }

            LIST_ADDTO_TAIL(&plocal->link, &pio_handle->psglh_list);
        }
        else
        {
            // Free all the SGL Handles
            hwi_free_chained_sgls(pcontroller->phwi_ws, pio_handle);
            *psgl_handle = ACIT_NULL;
            OSM_ASSERT_MSG(plocal, ("Freeing SGL handles during alloc!\n"));
            break;
        }
    }
}
#endif

ACIT_STATUS
hwi_alloc_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
#ifdef ACIT_CFG_CHAINED_SGL
    PSGL_HANDLE             psgl_handle = ACIT_NULL;
#endif

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    pio_handle->pwrb_handle = hwi_alloc_wrb_handle(phwi_controller, 
                                                   pio_handle->cri);
    if(pio_handle->pwrb_handle)
    {
        pio_handle->pwrb_handle->type = HWH_TYPE_IO;
        pio_handle->pwrb_handle->pio_handle = pio_handle;

#ifdef ACIT_CFG_CHAINED_SGL
        hwi_alloc_chained_sgls(pcontroller, pio_handle, &psgl_handle);

        if(psgl_handle)
        {
            pio_handle->itt = (((ACIT_UEXACT32)(pio_handle->pwrb_handle->wrb_index) << 16) |
                                (ACIT_UEXACT32)(psgl_handle->sgl_index));

            return ACIT_STATUS_SUCCESS;
        }
       
        // Free WRB handle before return
        hwi_free_wrb_handle(phwi_controller,
                            pio_handle->cri,
                            pio_handle->pwrb_handle);

#else
        pio_handle->itt = (((ACIT_UEXACT32)(pio_handle->pwrb_handle->wrb_index) << 16) |
                            (ACIT_UEXACT32)(pio_handle->psgl_handle->sgl_index));

        return ACIT_STATUS_SUCCESS;
#endif
    }

    return ACIT_STATUS_INSUFFICIENT_RESOURCE;
}

ACIT_VOID
hwi_free_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_free_wrb_handle(phwi_controller, 
                        pio_handle->cri, 
                        pio_handle->pwrb_handle);

    pio_handle->pwrb_handle->type = HWH_TYPE_INVALID;
    pio_handle->pwrb_handle->pio_handle = ACIT_NULL;

    pio_handle->pwrb_handle = ACIT_NULL;
#ifdef ACIT_CFG_CHAINED_SGL
    hwi_free_chained_sgls(pcontroller->phwi_ws, pio_handle);
#endif
}

ACIT_STATUS
hwi_alloc_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_alloc_eh_resources(phwi_controller, 
                           plogout_handle->cri,
                           &plogout_handle->pwrb_handle,
                           &plogout_handle->psgl_handle);

    plogout_handle->pwrb_handle->type = HWH_TYPE_LOGOUT;
    plogout_handle->pwrb_handle->plogout_handle = plogout_handle;

    plogout_handle->itt = (((ACIT_UEXACT32)(plogout_handle->pwrb_handle->wrb_index) << 16) |
                            (ACIT_UEXACT32)(plogout_handle->psgl_handle->sgl_index));

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_free_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_free_eh_resources(phwi_controller,
                          plogout_handle->cri,
                          plogout_handle->pwrb_handle,
                          plogout_handle->psgl_handle);

    plogout_handle->pwrb_handle = ACIT_NULL;
    plogout_handle->psgl_handle = ACIT_NULL;
}


ACIT_STATUS
hwi_alloc_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_alloc_eh_resources(phwi_controller, 
                           ptmf_handle->cri,
                           &ptmf_handle->pwrb_handle,
                           &ptmf_handle->psgl_handle);

    ptmf_handle->pwrb_handle->type = HWH_TYPE_TMF;
    ptmf_handle->pwrb_handle->ptmf_handle = ptmf_handle;

    ptmf_handle->itt = (((ACIT_UEXACT32)(ptmf_handle->pwrb_handle->wrb_index) << 16) |
                         (ACIT_UEXACT32)(ptmf_handle->psgl_handle->sgl_index));

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_free_tmf_resources(
    PACIT_CONTROLLER_WS         pcontroller,
    PACIT_TMF_HANDLE            ptmf_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    hwi_free_eh_resources(phwi_controller,
                         ptmf_handle->cri,
                         ptmf_handle->pwrb_handle,
                         ptmf_handle->psgl_handle);

    ptmf_handle->pwrb_handle = ACIT_NULL;
    ptmf_handle->psgl_handle = ACIT_NULL;
}

ACIT_UEXACT32
hwi_alloc_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle,
    ACIT_UEXACT32           cri,
    PACIT_NOP_HANDLE        poptional_nop_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    ACIT_UEXACT32           status;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    //
    // Only if optional handle is present, then allocate
    // an SGL. Otherwise, ITT will be -1 and no SGL will
    // be required since it will complete in drvr_message
    // path
    //
    if(poptional_nop_handle)
    {
        status = hwi_alloc_eh_resources(phwi_controller, 
                                        cri,
                                        &pnop_handle->pwrb_handle,
                                        &pnop_handle->psgl_handle);

        if(status == ACIT_STATUS_SUCCESS)
        {
            pnop_handle->pwrb_handle->type = HWH_TYPE_NOP;
            pnop_handle->pwrb_handle->pnop_handle = pnop_handle;

            pnop_handle->itt = (((ACIT_UEXACT32)(pnop_handle->pwrb_handle->wrb_index) << 16) |
                                 (ACIT_UEXACT32)(pnop_handle->psgl_handle->sgl_index));
        }
        else
        {
            return status;
        }
    }
    else
    {
        pnop_handle->psgl_handle = ACIT_NULL;
        pnop_handle->pwrb_handle = hwi_alloc_wrb_handle(phwi_controller, cri);
        OSM_ASSERT(pnop_handle->pwrb_handle);
        if(pnop_handle->pwrb_handle)
        {
            OSM_ASSERT(phwi_controller->nop_icd);
            pnop_handle->pwrb_handle->type = HWH_TYPE_NOP;
            pnop_handle->pwrb_handle->pnop_handle = pnop_handle;
            pnop_handle->itt = (((ACIT_UEXACT32)(pnop_handle->pwrb_handle->wrb_index) << 16) |
                                 (ACIT_UEXACT32)(phwi_controller->nop_icd));
        }
        else
        {
            return ACIT_STATUS_INSUFFICIENT_RESOURCE;
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
hwi_free_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PWRB_HANDLE             pwrb_h;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    pwrb_h = pnop_handle->pwrb_handle;

    // to handle those cases where a drvr message has arrived yet
    LIST_REMOVE_ELEMENT(&pwrb_h->link);
    LIST_INIT_ELEMENT(&pwrb_h->link);

    if(pnop_handle->psgl_handle)
    {
        hwi_free_eh_resources(phwi_controller,
                              pnop_handle->cri,
                              pnop_handle->pwrb_handle,
                              pnop_handle->psgl_handle);
    }
    else
    {
        hwi_free_wrb_handle(phwi_controller, 
                            pnop_handle->cri, 
                            pnop_handle->pwrb_handle);

        pnop_handle->pwrb_handle->type = HWH_TYPE_INVALID;
        pnop_handle->pwrb_handle->pio_handle = ACIT_NULL;
    }

    pnop_handle->pwrb_handle = ACIT_NULL;
    pnop_handle->psgl_handle = ACIT_NULL;
}

PACIT_IO_HANDLE
hwi_io_handle_from_itt(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           cri, 
    ACIT_UEXACT32           itt)
{
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PHWI_CONTROLLER_WS      phwi;
    PWRB_HANDLE             pwrb_handle;
    ACIT_UEXACT16           wrb_index = 0, sgl_index = 0;
    ACIT_UEXACT32           icd_begin[BE_MAX_ULP];

    phwi = GET_HWI_CONTROLLER_WS(pc);

    wrb_index = (ACIT_UEXACT16)((itt & 0x00ff0000) >> 16);
    sgl_index = (ACIT_UEXACT16)(itt & 0x00000fff);

    icd_begin[0] = pc->fw_config.ulp[0].icd_start;
    icd_begin[1] = pc->fw_config.ulp[1].icd_start;

    if((wrb_index < pc->params.num_cmd_per_cxn) &&
        (((sgl_index >= icd_begin[0]) && (sgl_index < (icd_begin[0] + pc->params.icds_per_controller))) ||
         ((sgl_index >= icd_begin[1]) && (sgl_index < (icd_begin[1] + pc->params.icds_per_controller)))))
    {
        OSM_ASSERT(phwi->pwrb_context[cri].cid == hwi_cid_from_cri(pc, cri));

        pwrb_handle = (PWRB_HANDLE)((PACIT_UEXACT8)phwi->pwrb_context[cri].\
                        pwrb_handle_base + wrb_index * sizeof(WRB_HANDLE));

        if(pwrb_handle->type == HWH_TYPE_IO)
        {
            return pwrb_handle->pio_handle;
        }
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("ITT 0x%x - WRB 0x%x SGL 0x%x not found cri = %x\n", 
                                        itt, wrb_index, sgl_index, cri));
    OSM_ASSERT(0);
    return ACIT_NULL;
}


PBE_FUNCTION_OBJECT
hwi_get_be_pfo(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc;
    PHWI_CONTEXT_MEMORY     phwi_context;
    PBE_FUNCTION_OBJECT     pfo;

    phwc = GET_HWI_CONTROLLER_WS(pcontroller);
    ASSERT (phwc);
    phwi_context = phwc->phwic;
    OSM_ASSERT(phwi_context);

    pfo = &phwi_context->be_function;

    return pfo;
}

ACIT_UEXACT32
hwi_fill_mcc_wrb_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len,
    PACIT_VOID              presponse_hdr,
    ACIT_UEXACT32           response_hdr_len,
    PACIT_VOID              posm_reference)
{
    ACIT_BUS_ADDRESS        pa;
    ACIT_SEXACT32           length, offset, index;

    if(payload_len)
    {
        length = payload_len;
        offset = 0;
        pwrb->sge_count = 0;

        while(length > 0)
        {
            if(pwrb->sge_count > BE_MAX_SGE_IN_MCC_WRB)
            {
                return ACIT_STATUS_ERROR;
            }

            pa = OSM_GET_BUS_ADDRESS(pcontroller->pcontroller_key,
                                     posm_reference,
                                     (PACIT_UEXACT8)ppayload + offset,
                                     (PACIT_UEXACT32)&length);

            index = pwrb->sge_count + 1;

            OSM_ASSERT(length && pa.u.a64.address);

            ACIT_PUT_LE32(pwrb->payload.sgl[index].pa_lo,
                          pa.u.a32.address_lo);
            ACIT_PUT_LE32(pwrb->payload.sgl[index].pa_hi,
                          pa.u.a32.address_hi);

            if(length <= (ACIT_SEXACT32)payload_len)
            {
                ACIT_PUT_LE32(pwrb->payload.sgl[index].length,
                              length);
            }
            else
            {
                ACIT_PUT_LE32(pwrb->payload.sgl[index].length,
                              payload_len);
            }

            offset += length;
            length = payload_len - offset;
            pwrb->sge_count++;

            if(length <= 0)
                break;
        }
    }

    if(response_hdr_len)
    {
        length = response_hdr_len;
        pa = OSM_GET_BUS_ADDRESS(pcontroller->pcontroller_key,
                                    posm_reference,
                                    (PACIT_UEXACT8)presponse_hdr,
                                    (PACIT_UEXACT32)&length);

        OSM_ASSERT(length && pa.u.a64.address);

        ACIT_PUT_LE32(pwrb->payload.sgl[0].pa_lo,
                        pa.u.a32.address_lo);
        ACIT_PUT_LE32(pwrb->payload.sgl[0].pa_hi,
                        pa.u.a32.address_hi);

        if(length <= (ACIT_SEXACT32)response_hdr_len)
        {
            ACIT_PUT_LE32(pwrb->payload.sgl[0].length,
                          length);
        }
        else
        {
            ACIT_PUT_LE32(pwrb->payload.sgl[0].length,
                          response_hdr_len);
        }

        pwrb->sge_count++;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
hwi_fill_non_embedded_wrb(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len,
    PACIT_VOID              presponse_hdr,
    ACIT_UEXACT32           response_hdr_len,
    PACIT_VOID              posm_reference)
{
    ACIT_STATUS             status;

    /*
     * For non embedded WRBs, the payload length should also contain
     * the IOCTL header length
     */
    ACIT_PUT_LE32(pwrb->payload_length, payload_len + response_hdr_len);

    /*
     * Construct a SGL for the payload and fill them into
     * MCC WRB
     */

    pwrb->embedded = 0;

    status = hwi_fill_mcc_wrb_sgl(pcontroller,
                                    pwrb,
                                    ppayload,
                                    payload_len,
                                    presponse_hdr,
                                    response_hdr_len,
                                    posm_reference);

    return status;
}
ACIT_STATUS
hwi_fill_embedded_wrb(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len)
{

    /*
     * Directly copy the payload into the MCC WRB. This is an embedded
     * MCC WRB since there's no response header. The CQ entry is good
     * enough to return the status
     */

    pwrb->embedded = 1;

    return ACIT_STATUS_SUCCESS;
}

#if 0
// STORPORT_DUPLEX not implemented

ACIT_UEXACT32
hwi_check_wrb_index(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT16           cri, 
    ACIT_UEXACT16           wrb_index)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PISCSI_CQE              pcqe;
    SA_RING                 *ring;
    ACIT_U32                entries, found = 0;
    ACIT_UEXACT32           cid = 0;
 
    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &phwi_controller->cq;
    pcqe = (PISCSI_CQE) ring->va;
    cid = hwi_cid_from_cri(pcontroller, cri);

    for(entries = 0; entries < ring->num; entries++)
    {
        if(pcqe->valid)
        {
            // compare CID and WRB index
            if(cid == pcqe->cid)
            {
                PI_SOL_CQE psol = (PI_SOL_CQE)pcqe;

                if((psol->code == SOL_CMD_COMPLETE) &&
                  (wrb_index == psol->wrb_index))
                {
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Found at VA 0x%p - CQ entry:", pcqe));
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("0x%x 0x%x 0x%x 0x%x\n", pcqe->dw[0], pcqe->dw[1], pcqe->dw[2], pcqe->dw[3]));
                    found = 1;
                }
            }
        }
        pcqe++;
    }

    return found;
}

ACIT_UEXACT32
hwi_check_pending_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid,
    ACIT_UEXACT32           wrb_index)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PWRB_HANDLE             pwrb_handle;
    ACIT_UEXACT32           found;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);

    pwrb_handle = (PWRB_HANDLE)(phwi_controller->pwrb_context[hwi_cri_from_cid(pcontroller, cid)].pwrb_handle_base) +
                                wrb_index;

    if(pwrb_handle->type != HWH_TYPE_INVALID)
        found = 1;
    else
        found = 0;

    return found;
}

ACIT_VOID
hwi_check_stale_cq(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwi_controller;
    PISCSI_CQE              pcqe;
    SA_RING                 *ring;
    ACIT_UEXACT32           entries, found = 0, stales = 0;
    PI_SOL_CQE              psol;

    phwi_controller = GET_HWI_CONTROLLER_WS(pcontroller);
    ring = &phwi_controller->cq;
    pcqe = (PISCSI_CQE) ring->va;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Scanning CQEs for stale entries\n"));

    for(entries = 0; entries < ring->num; entries++)
    {
        if(pcqe->valid && pcqe->code == SOL_CMD_COMPLETE)
        {
            psol = (PI_SOL_CQE)pcqe;

            found = hwi_check_pending_handle(pcontroller, 
                                             psol->cid, 
                                             psol->wrb_index);

            if(!found)
            {
                stales++;
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("!!!!!Entry at VA 0x%p stale!! - CQ entry:\n", pcqe));
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("0x%x 0x%x 0x%x 0x%x\n", pcqe->dw[0], pcqe->dw[1], pcqe->dw[2], pcqe->dw[3]));

            }
        }
        pcqe++;
    }

    if(!stales)
	{
       	 OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("No stale entries found in CQ\n"));
	}
    else
	{
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("%d stale entries found in CQ\n", stales));
        OSM_ASSERT(0);
	}
}

ACIT_VOID
hwi_validate_pending_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget_to_search,
    PACIT_LUN_WS            plun_to_search)
{
    ACIT_UEXACT32           target, found = 0;
    PACIT_TARGET_WS         ptarget;
    PACIT_DC_LIST           flink, pio_link;
    PACIT_IO_HANDLE         pio_handle;
    PACIT_TARGET_ENTRY      ptarget_entry = ACIT_NULL;
    PACIT_IO_HANDLE         pioh_oldest_t, pioh_oldest_c = ACIT_NULL;
    ACIT_UEXACT32           num_ios_t = 0, num_ios_c = 0, isread;
    ACIT_UEXACT32           num_r_ios_t = 0, num_r_ios_c = 0, num_w_ios_t = 0, num_w_ios_c = 0;

    //
    // First scan the pending list for each target and compare
    // the CQ to see if there is an entry
    //
    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(target = 0; target < pcontroller->ptarget_table->occupied_entries; target++)
    {
        ptarget = GET_TARGET_WS_SAFE(ptarget_entry[target].ptarget_key);

        if(!ptarget)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Ignoring deleted or not present target..\n"));
            continue;
        }

        if(ptarget_to_search && (ptarget_to_search != ptarget))
            continue;

        pioh_oldest_t = ACIT_NULL;
        num_ios_t = 0;
        num_r_ios_t = 0;
        num_w_ios_t = 0;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Scanning target %d ws 0x%p...\n", target, ptarget));

        flink = ptarget->io_handle_busy_list.next;
        pio_link = ptarget->io_handle_busy_list.next;

        while(flink != &ptarget->io_handle_busy_list)
        {
            pio_handle = STRUCT_BASE(ACIT_IO_HANDLE,
                                     target_link,
                                     pio_link);

            OSM_ASSERT(pio_handle);

            flink = flink->next;
            pio_link = pio_link->next;

            if(plun_to_search && (plun_to_search != pio_handle->plun))
                continue;

            if(pio_handle->pio->pio_request->flags.inbound_data)
            {
                isread = 1;
                num_r_ios_t++;
            }
            else
            {
                isread = 0;
                num_w_ios_t++;
            }

            found = hwi_check_wrb_index(pcontroller, 
                                        pio_handle->cri, 
                                        (pio_handle->itt >> 16) & 0xffff);

            if(!found)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,(("Missing CEV on IO Handle 0x%p Cid 0x%x CmdSN 0x%x ITT 0x%x Read=%d TS 0x%x\n"), 
                                                            (PACIT_UEXACT32)pio_handle, 
                                                            hwi_cid_from_cri(pcontroller, pio_handle->cri),
                                                            pio_handle->cmd_sn,
                                                            pio_handle->itt,
                                                            isread,
                                                            pio_handle->starting_tick_count));
            }

            if(!pioh_oldest_t)
                pioh_oldest_t = pio_handle;
            else 
                if(pio_handle->starting_tick_count < pioh_oldest_t->starting_tick_count)
                    pioh_oldest_t = pio_handle;

            num_ios_t++;
        }

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Summary for target %d\n", target));
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Pending IOs %d Reads %d Writes %d\n", num_ios_t, num_r_ios_t, num_w_ios_t));
        if(num_ios_t)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Oldest IOH 0x%p TS 0x%x\n", pioh_oldest_t, pioh_oldest_t->starting_tick_count));
        }

        if(num_w_ios_t > 32)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("!!DRIVER ERROR - %d WRITES OUTSTANDING!!\n",num_w_ios_t));
        }

        if(!pioh_oldest_c)
            pioh_oldest_c = pioh_oldest_t;
        else 
            if((pioh_oldest_t) && 
               (pioh_oldest_t->starting_tick_count < pioh_oldest_c->starting_tick_count))
                    pioh_oldest_c = pioh_oldest_t;

        num_ios_c += num_ios_t;
        num_r_ios_c += num_r_ios_t;
        num_w_ios_c += num_w_ios_t;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Summary for controller\n"));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Pending IOs %d Reads %d Writes %d\n", num_ios_c, num_r_ios_c, num_w_ios_c));
    if(num_ios_c)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, (("Oldest IOH 0x%p TS 0x%x Current TS 0x%x\n"), 
                                            pioh_oldest_c, 
                                            pioh_oldest_c->starting_tick_count,
                                            CURRENT_TICK(pcontroller)));
    }

    //
    // Then do the reverse. Go through the CQ first and try and 
    // account for each entry in controller's list. Not sure how useful 
    // it is, but may catch some stale CQ entries
    //
    hwi_check_stale_cq(pcontroller);

}
#endif

ACIT_VOID
hwi_get_pci_link_info(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PCICFG_PCIE_LINK_STATUS_CSR pcie_link_status;
    PCICFG_PCIE_LINK_CAP_CSR pcie_link_cap;
    UEXACT32 log[2];

    pcie_link_status.dw = ACIT_PCICFG0_READ(pcontroller, pcie_link_status);
    pcie_link_cap.dw = ACIT_PCICFG0_READ(pcontroller, pcie_link_cap);

    pcontroller->pcie_info.current_link_width = (ACIT_UEXACT8)pcie_link_status.NegLinkWidth;
    pcontroller->pcie_info.max_link_width = (ACIT_UEXACT8)pcie_link_cap.MaxLinkWidth;

    pcontroller->pcie_info.current_link_speed = (ACIT_UEXACT8)pcie_link_status.LinkSpeed;
    pcontroller->pcie_info.max_link_speed =  (ACIT_UEXACT8)pcie_link_cap.MaxLinkSpeed;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, (("LinkWidth - Current/Max=%d/%d LinkSpeed - Current/Max=%d/%d\n"), 
                                            pcontroller->pcie_info.current_link_width, 
                                            pcontroller->pcie_info.max_link_width,
                                            pcontroller->pcie_info.current_link_speed,
                                            pcontroller->pcie_info.max_link_speed));

    log[0] = pcontroller->pcie_info.current_link_speed;
    log[1] = pcontroller->pcie_info.current_link_width;

    ap_log_osm_event(pcontroller,
                    UNI_CFG_LOG_LEVEL_INFO,
                    UNI_SUBMODULE_ACIT_HWI,
                    UNI_PHASE_INIT,
                    2,
                    ACIT_EC_HWI_PCIE_INFO,
                    0, ACIT_NULL,
                    &log[0]);

    return;
}

/*
ACIT_VOID
hwi_get_pci_ids(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PCICFG_ID_F0_CSR pci_id;

    pci_id.dw = ACIT_PCICFG0_READ(pcontroller, id);

    pcontroller->vendor_id = (ACIT_UEXACT16)pci_id.vendorid;
    pcontroller->device_id = (ACIT_UEXACT16)pci_id.deviceid;

    return;
}
*/

ACIT_VOID
hwi_cleanup_after_ue(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc = NULL;
    PHWI_CONTEXT_MEMORY     phwi_context;

    //
    // This routine will be called after a UE has been detected. Call the
    // beclib abort routine to complete all pending IOCTLs with beclib.
    //

    phwc = GET_HWI_CONTROLLER_WS (pcontroller);
    phwi_context = phwc->phwic;

    be_function_abort_ioctls(&phwi_context->be_function);
 }

ACIT_UEXACT32
hwi_start_ue_recovery(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc = NULL;
    PHWI_CONTEXT_MEMORY     phwi_context;
    BESTATUS                bestatus;

    //
    // This routine will be called after a UE has been detected and a
    // UE recovery needs to be initiated. As per the beclib documentation
    // this can run from DISPATCH level.
    //

    phwc = GET_HWI_CONTROLLER_WS (pcontroller);
    phwi_context = phwc->phwic;

    bestatus = be_ue_recovery_start(&phwi_context->be_error_obj, 
                                    0);   // not crashdump

    if(bestatus == BE_SUCCESS)
        return ACIT_STATUS_SUCCESS;
    else
        return ACIT_STATUS_ERROR;
}

ACIT_UEXACT32
hwi_poll_for_ue_recovery(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PHWI_CONTROLLER_WS      phwc = NULL;
    PHWI_CONTEXT_MEMORY     phwi_context;
    BESTATUS                bestatus;

    //
    // This routine will be called after a UE has been detected and a
    // UE recovery has been initiated.
    //

    phwc = GET_HWI_CONTROLLER_WS (pcontroller);
    phwi_context = phwc->phwic;

    bestatus = be_ue_recovery_poll(&phwi_context->be_error_obj);

    switch(bestatus)
    {
    case BE_PENDING:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("UE Recovery Pending...\n"));
        return ACIT_STATUS_PENDING;

    case BE_SUCCESS:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("****UE Recovery Success****\n"));
        return ACIT_STATUS_SUCCESS;

    default:
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("****UE Recovery Error****%x\n", bestatus));
        return ACIT_STATUS_ERROR;
    }
}

ACIT_UEXACT32
hwi_init_cid_map(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           num_elements)
{
    PHWI_CONTROLLER_WS      phwc = NULL;
    ACIT_UEXACT32           cri, mapidx;

    //
    // This routine will be called to initialize the CidMap table. This
    // routine MUST be called after all the WRB rings have been created
    // for the function.
    //

    phwc = GET_HWI_CONTROLLER_WS (pcontroller);

    // Check if our allocation is large enough and if not, we fail
    // the routine.
    if(num_elements > CID_MAP_SZ)
    {
        return ACIT_STATUS_INSUFFICIENT_RESOURCE;
    }

    //
    // Initialize the cid_map table.
    // Look for CID_MAP_SZ definition for documentation on the CID map
    // table.
    for(cri=0; cri<num_elements; cri++)
    {
        mapidx = phwc->pwrb_context[cri].cid;
        phwc->cid_map.map[mapidx] = (ACIT_UEXACT16)cri;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT16
hwi_cid_from_cri(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cri)
{
    PHWI_CONTROLLER_WS      phwic;
    phwic = GET_HWI_CONTROLLER_WS(pcontroller);

    // No checking here!
    OSM_ASSERT(cri < BE_MAX_CXNS);

    return (ACIT_UEXACT16)phwic->pwrb_context[cri].cid;
}

ACIT_UEXACT16
hwi_cri_from_cid(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid)
{
    PHWI_CONTROLLER_WS      phwic;
    phwic = GET_HWI_CONTROLLER_WS(pcontroller);

    // No checking here!
    OSM_ASSERT(cid < CID_MAP_SZ);

    // There are enough array elements allocated
    return phwic->cid_map.map[cid];
}

ACIT_UEXACT32
_hwi_get_ulp_for_cri(
    PHWI_CONTROLLER_WS      phwi_controller,
    ACIT_UEXACT32           cri)
{
    return phwi_controller->pwrb_context[cri].ulp;
}

ACIT_UEXACT32
hwi_get_ulp_for_cri(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cri)
{
    PHWI_CONTROLLER_WS      phwic;
    phwic = GET_HWI_CONTROLLER_WS(pcontroller);

    return _hwi_get_ulp_for_cri(phwic, cri);
}
