/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        hwinit.c

    Purpose:

        This file implements the subroutines that are used to initializate
        the ASIC.
        

--*/


#include "acitpch.h"

ACIT_UEXACT32
hwi_init_mem_bars(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PACIT_VOID              osm_handle)
{
    SA_DEV_BAR_LOCATIONS    bars[3];
    ACIT_UEXACT32 err_status = 0;
    BESTATUS bestatus;

    bars[0].base_pa = pc->bar0_pa.u.a64.address;
    bars[0].base_va = pc->bar0_va;
    bars[0].length = sizeof (PCICFG0_CSRMAP);
    bars[0].mem_or_io_mapped = SA_MEM_MAPPED;
    bars[0].type = SA_BAR_TYPE_PCI;
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,
                        "BAR0",
                        bars[0].base_va);

    bars[1].base_pa = pc->bar1_pa.u.a64.address;
    bars[1].base_va = pc->bar1_va;
    bars[1].length = sizeof (BLADE_ENGINE_CSRMAP);
    bars[1].mem_or_io_mapped = SA_MEM_MAPPED;
    bars[1].type = SA_BAR_TYPE_CSR;
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,
                        "BAR1",
                        bars[1].base_va);

    bars[2].base_pa = pc->bar2_pa.u.a64.address;
    bars[2].base_va = pc->bar2_va;
    bars[2].length = sizeof (PROTECTION_DOMAIN_DBMAP);
    bars[2].mem_or_io_mapped = SA_MEM_MAPPED;
    bars[2].type = SA_BAR_TYPE_PD;
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,
                        "BAR2",
                        bars[2].base_va);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Creating SA Device..\n"));

    bestatus = (BESTATUS)sa_dev_create (bars, 3, osm_handle, &pcontext->sa_dev);

    if(bestatus != BE_SUCCESS)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("sa_dev_create failed - %d\n", bestatus));
        err_status = HWI_ERR_SA_DEV_CREATE;
    }

    return err_status;
}

ACIT_UEXACT32
hwi_init_belib_library(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    SA_SGL sgl = {0}, *pemul = NULL;
#ifdef BE_EMULATION
    SA_SGL emul_sgl = {0};
#endif
    ACIT_UEXACT32 err_status = 0;
    BESTATUS bestatus;

    bestatus = be_initialize_library ();
    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "Initialize beclib library failed.");
        err_status = HWI_ERR_BECLIB_INIT;
        goto error;
    }

    pcontext->initialized.lib_init = 1;

     // Let the lib create its internal objects...
    hwi_build_sa_sgl (pc, HWI_MEM_MAILBOX, &sgl);

#ifdef BE_EMULATION
    hwi_build_sa_sgl (pc, HWI_MEM_EMUL, &emul_sgl);
    pemul = &emul_sgl;
#endif

    bestatus = be_function_and_chip_create (
                        &pcontext->sa_dev,
                        BE_FUNCTION_TYPE_ISCSI,
                        &sgl,
                        pemul,
                        pfo,
                        &pcontext->be_chip); // object to initialize

    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "function create failed.");
        err_status = HWI_ERR_FUN_CHIP_CREATE;
        goto error;
    }

    pcontext->initialized.chip_function = 1;

    if(pcontext->initialized.ue_object)
    {
        // Link the function object and error object.
        be_function_link_error_object(pfo, &pcontext->be_error_obj);
    }

    return ACIT_STATUS_SUCCESS;

error:
    return err_status;
}

ACIT_UEXACT32
hwi_check_ue_at_init(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext)
{
    ACIT_UEXACT32 err_status = 0;
    BESTATUS bestatus;

    if(pcontext->initialized.ue_object == 0)
    {
        bestatus = be_ue_create (&pcontext->sa_dev, &pcontext->be_error_obj);

        if(bestatus != BE_SUCCESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("be_ue_create failed - %d\n", bestatus));
            err_status = HWI_ERR_UE_OBJ_CREATE;
            goto error;
        }

        pcontext->initialized.ue_object = 1;
    }

    // An alternative implementation is to use be_ue_check_hardware_ue. But that does not
    // let us know what the UE is. Hence use our own routine..
    //bestatus = be_ue_check_hardware_ue(&phwi->phwic->be_error_obj);

    if(ACIT_STATUS_SUCCESS != hwi_check_and_log_ue(pc))
    {
        ACIT_U32 recovered = 0;

        while(ACIT_IS_UE_RECOVERY_ENABLED(pc))
        {
            // Initiate UE Recovery
            bestatus = be_ue_recovery(&pcontext->be_error_obj,
                (pc->state.init_type == ACIT_INIT_TYPE_CRASHDUMP) ? 1 : 0);

            // Increment the UE recoveries..
            pc->eh_stats.ue_recoveries++;

            if(bestatus == BE_SUCCESS)
            {
                // Successfully recovered from an UE.
                ap_flag_ue_recovery_complete(pc);
                recovered = 1;
                break;
            }
        }

        if(recovered == 0)
        {
            // Log an event to indicate UE tolerance
            ap_log_osm_event(pc,
                             UNI_CFG_LOG_LEVEL_ERROR,
                             UNI_SUBMODULE_ACIT_HWI,
                             UNI_PHASE_DONT_CARE,
                             0,
                             ACIT_EC_HWI_IN_UE_TOLERANCE,
                             0, ACIT_NULL,
                             ACIT_NULL);

            err_status = HWI_ERR_UE_AT_INIT;
            goto error;
        }
    }

error:
    return err_status;
}

ACIT_UEXACT32
hwi_check_and_issue_reset(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    ACIT_UEXACT32 err_status = 0;
    MGMT_STATUS mgmt_status;
    BESTATUS bestatus;

    // A reset is needed only if we are in crashdump
    if(pc->state.init_type == ACIT_INIT_TYPE_CRASHDUMP)
    {
        if(pc->params.crash_dump.dump_rst == 1)
        {
            if(pc->state.soft_reset == 0)
            {
                bestatus = be_pci_soft_reset(pfo);

                if (bestatus != BE_SUCCESS)
                {
                    TRACE (DL_ERR, "soft reset failed.");
                    err_status = HWI_ERR_SOFT_RESET;
                    goto error;
                }

                pc->state.soft_reset = 1;
                be_function_and_chip_destroy (pfo, &pcontext->be_chip); // object to destroy
                pcontext->initialized.chip_function = 0;
            }
        }
        else
        {
            //
            // Bug 18349: A soft reset has some issues on some platforms where the CLP configured
            // data is not on the flash. So a reset will wipe out this info and ARM fw will not
            // be able to establish session to boot target. So avoid the chip reset and go with
            // function level reset.
            // FLR will work only when there's no UE. So check for UE first and then issue FLR 
            // IOCTL.
            //
            // Note!!
            // There's still a way to trigger a chip reset. Parameter dump_rst must be configured 
            // in the registry/insmod time and must be passed to the dump driver
            //
            hwi_check_health(pc, ACIT_FALSE);

            if(pc->eh_stats.ue_detected == 0)
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Issuing function level reset..\n"));
                mgmt_status = mgmt_issue_function_reset(pc);

                if(mgmt_status != MGMT_STATUS_SUCCESS)
                {
                    bestatus = BE_NOT_OK;
                    err_status = HWI_ERR_FUNCTION_RESET;
                    TRACE (DL_ERR, "Failed to issue Function Reset, quitting\n");
                    goto error;
                }
                else
                {
                    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Function level reset completed successfully\n"));
                }
            }
        }
    }

error:
    return err_status;
}

ACIT_UEXACT32
hwi_init_sa_rings(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 pid, num_eq_entries, num_cq_entries;
    PACIT_VOID eq_va, cq_va;
    ACIT_UEXACT32 err_status = 0, ulp = 0;

    // Create SA rings for CQ, EQ and RQ

    eq_va = (PACIT_UEXACT8)GET_MEM_DESC_VA(pc, HWI_MEM_EQ);

    num_eq_entries = (p->num_eq_pages * PAGE_SIZE)/sizeof (EQ_ENTRY);
    num_cq_entries = (p->num_cq_pages * PAGE_SIZE)/sizeof (ISCSI_CQE);

    for(pid = 0; pid < p->num_cq; pid++)
    {
        cq_va = (PACIT_UEXACT8)GET_MEM_DESC_VA(pc, phwi->cq_mem_index + (pid/8));

        sa_ring_create(&pc->partition[pid].eq,
                num_eq_entries, // num entries
                sizeof (EQ_ENTRY),
                (PACIT_UEXACT8)eq_va + pid * num_eq_entries * sizeof(EQ_ENTRY));

        // the mod 8 is needed because it's the offset from current VA, not
        // the base VA like the EQ ring

        sa_ring_create(&pc->partition[pid].cq,
                num_cq_entries, // num entries
                sizeof (ISCSI_CQE),
                (PACIT_UEXACT8)cq_va + (pid%8) * num_cq_entries * sizeof(ISCSI_CQE));
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                    ("EQ VA - 0x%p\n", pc->mem_descriptors[HWI_MEM_EQ].virtual_address));
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                    ("CQ VA - 0x%p\n", pc->mem_descriptors[phwi->cq_mem_index].virtual_address));

    for(ulp = 0; ulp < BE_MAX_ULP; ulp++)
    {
        if(p->ulp_mask & (1 << ulp))
        {
            sa_ring_create(&phwi->default_pdu_hdr[ulp],
                      GET_MEM_DESC_SZ(pc, HWI_MEM_ASYNC_HEADER0_RING+ulp)/ sizeof (PHYS_ADDR), // num entries
                      sizeof (PHYS_ADDR),
                      pc->mem_descriptors[HWI_MEM_ASYNC_HEADER0_RING+ulp].virtual_address);

            sa_ring_create(&phwi->default_pdu_buf[ulp],
                      GET_MEM_DESC_SZ(pc, HWI_MEM_ASYNC_DATA0_RING+ulp)/ sizeof (PHYS_ADDR), // num entries
                      sizeof (PHYS_ADDR),
                      pc->mem_descriptors[HWI_MEM_ASYNC_DATA0_RING+ulp].virtual_address);

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                ("RQ-Hdr[%d]  VA - 0x%p\n",
                ulp, pc->mem_descriptors[HWI_MEM_ASYNC_HEADER0_RING+ulp].virtual_address));

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                ("RQ-Dta[%d] VA - 0x%p\n",
                ulp, pc->mem_descriptors[HWI_MEM_ASYNC_DATA0_RING+ulp].virtual_address));
        }
    }

    return err_status;
}

ACIT_UEXACT32
hwi_init_eq(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 timer_threshold;
    ACIT_UEXACT32 pid, err_status = 0;
    BESTATUS bestatus;
    SA_SGL sgl = {0};

    // config params is in usecs and the units need to be in 8 usec units
    timer_threshold = p->eq_timer / 8;
    if(timer_threshold > CEV_TIMER_MAX)
        timer_threshold = CEV_TIMER_MAX;

    for(pid = 0; pid < p->num_cq; pid++)
    {
        hwi_build_sa_sgl_by_offset (pc,
                                    HWI_MEM_EQ,
                                    (pid * p->num_eq_pages * PAGE_SIZE),
                                    p->num_eq_pages * PAGE_SIZE,
                                    &sgl);

        bestatus = be_eq_create (pfo,
                                &sgl,
                                sizeof (EQ_ENTRY),
                                p->num_eq_entries,
                                be_eq_delay_to_multiplier(pfo, timer_threshold), // timer delay in 8 usec units
                                &pcontext->phwi_queue[pid].be_eq);

        if (bestatus != BE_SUCCESS)
        {
            TRACE (DL_ERR, "eq create failed.");
            err_status = HWI_ERR_EQ_CREATE;
            goto error;
        }

        pc->partition[pid].eq.id = be_eq_get_id(&pcontext->phwi_queue[pid].be_eq);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("EQ created with RID %d\n", pc->partition[pid].eq.id));

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("EQ Object : 0x%p\n", &pcontext->phwi_queue[pid].be_eq));
    }

    pcontext->initialized.eq = 1;

error:
    return err_status;
}

ACIT_UEXACT32
hwi_init_mgmt_attrib(
    PACIT_CONTROLLER_WS     pc)
{
#if defined (BE_EMULATION) || defined (FPGA_ARM)
    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Emulation or FPGA build, skipping fw_config IOCTL\n"));
    return 0;
#else
    MGMT_STATUS mgmt_status;
    ACIT_UEXACT32 err_status = 0;

    // Check the firmware version to see if we support it
    mgmt_status = mgmt_check_supported_fw(pc);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        TRACE (DL_ERR, "Unsupported fw version\n");
        err_status = HWI_ERR_SUPPORTED_FW_CHECK;
        goto error;
    }

    // Fetch and store fw config
    mgmt_status = mgmt_get_fw_config(pc);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        TRACE (DL_ERR, "Error getting fw config\n");
        err_status = HWI_ERR_FW_CONFIG_GET;
        goto error;
    }

    // Fetch and store management attributes
    mgmt_status = mgmt_get_hba_attribs(pc);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        TRACE (DL_ERR, "Error getting hba attribs\n");
        err_status = HWI_ERR_HBA_ATTRIB_GET;
        goto error;
    }

    // Check if interrupt redirection is supported by this FW
    mgmt_status = mgmt_get_addi_attribs(pc);

    if(mgmt_status != MGMT_STATUS_SUCCESS)
    {
        //
        // Ideally the IOCTL should not have failed. Lets continue with
        // the initialization anyway..
        //
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Failed to query additional attribs!\n"));
        pc->addi_attribs.interrupt_redirection = 0;
        OSM_ASSERT(0);
    }

error:
    return err_status;
#endif

}

ACIT_UEXACT32
hwi_init_mcc_rings(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 err_status = 0;
    BESTATUS bestatus;
    SA_SGL sgl = {0};

    hwi_build_sa_sgl (pc, HWI_MEM_MCC_CQ, &sgl);
    bestatus = be_cq_create (pfo, 
                             &sgl, 
                             sa_sgl_get_byte_count (&sgl), // num entries
                             ACIT_FALSE,                   // not eventable
                             ACIT_TRUE,                    // force interrupt
                             BE_CQ_DMA_COALESCE_NONE,      // DMA coalescing
                             &pcontext->phwi_queue[0].be_eq, // associated event queue object
                             &pcontext->be_mcc_cq);

    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "mcc cq create failed.");
        err_status = HWI_ERR_MCC_CQ_CREATE;
        goto error;
    }
    pcontext->initialized.mcc_cq = 1;

    OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
        ("MCC CQ created with RID %d\n", be_cq_get_id(&pcontext->be_mcc_cq)));

    // Create MCC send ring, which also transitions the chip to MCC mode...
    hwi_build_sa_sgl (pc, HWI_MEM_MCC, &sgl);
    bestatus = be_mcc_ring_create (pfo, 
                                   &sgl, 
                                   sa_sgl_get_byte_count (&sgl), 
                                   pcontext->pbe_mcc_context, 
                                   p->ioctls_per_ctrl,
                                   &pcontext->be_mcc_cq, 
                                   &pcontext->be_mcc);
    if (bestatus != BE_SUCCESS)
    {
        TRACE (DL_ERR, "mcc create failed.");
        err_status = HWI_ERR_MCC_RING_CREATE;
        goto error;
    }
    pcontext->initialized.mcc = 1;

error:
    return err_status;
}

ACIT_UEXACT32
hwi_init_cq(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 err_status = 0, pid;
    BESTATUS bestatus;
    SA_SGL sgl = {0};

    for(pid = 0; pid < p->num_cq; pid++)
    {
        // the mod 8 is needed because it's the offset from current VA, not
        // the base VA like the EQ ring

        hwi_build_sa_sgl_by_offset (pc, 
                                    phwi->cq_mem_index + (pid / 8), 
                                    ((pid%8) * p->num_cq_pages * PAGE_SIZE),
                                    p->num_cq_pages * PAGE_SIZE,
                                    &sgl);

        if(IS_ASIC_BE2_OR_BE3(pc))
        {
            bestatus = be_cq_create (pfo, 
                                    &sgl, 
                                    sa_sgl_get_byte_count (&sgl), 
                                    ACIT_FALSE,    // not eventable
                                    ACIT_FALSE,    // honor EQ timers
                                    BE_CQ_DMA_COALESCE_NONE, // DMA coalescing
                                    &pcontext->phwi_queue[pid].be_eq, // associated event queue object
                                    &pcontext->phwi_queue[pid].be_cq);
        }
        else
        {
            bestatus = be_cq_create_v2 (pfo, 
                                    &sgl, 
                                    sa_sgl_get_byte_count (&sgl), 
                                    ACIT_FALSE,    // honor EQ timers
                                    BE_CQ_DMA_COALESCE_NONE, // DMA coalescing
                                    &pcontext->phwi_queue[pid].be_eq, // associated event queue object
                                    &pcontext->phwi_queue[pid].be_cq,
                                    CQ_TYPE_DEFAULT,
                                    0, // autovalid doesn't apply for BE4
                                    ACIT_NULL, // callback
                                    ACIT_NULL, // callback context
                                    ACIT_NULL); // queue context
        }

        if (bestatus != BE_SUCCESS)
        {
            TRACE (DL_ERR, "cq create failed.");
            err_status = HWI_ERR_ISCSI_CQ_CREATE;
            goto error;
        }

        pc->partition[pid].cq.id = be_cq_get_id(&pcontext->phwi_queue[pid].be_cq);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("iSCSI CQ created with RID %d\n", pc->partition[pid].cq.id));

        pcontext->initialized.cq = 1;
    }

error:
    return err_status;
}

ACIT_UEXACT32
_hwi_init_rq(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 err_status = 0, i, def_pdu_ring_sz;
    BESTATUS bestatus;
    SA_SGL sgl[2] = {0};

    def_pdu_ring_sz = (p->num_async_pdus * sizeof(PHYS_ADDR));
    hwi_build_sa_sgl (pc, HWI_MEM_ASYNC_HEADER0_RING, &sgl[0]);
    hwi_build_sa_sgl (pc, HWI_MEM_ASYNC_DATA0_RING, &sgl[1]);

    // The i iterates through header and data
    for(i=0; i<2; i++)
    {
        bestatus = be_iscsi_defq_create (pfo,
                                         &sgl[i],
                                         def_pdu_ring_sz, // length
                                         p->default_pdu_buf_sz,
                                         &pcontext->phwi_queue[0].be_cq,
                                         &pcontext->be_defq[BE_ULP0_NUM_SLI][i]);

        if (bestatus != BE_SUCCESS)
        {
            TRACE (DL_ERR, "defq create failed.");
            err_status = 1;
            break;
        }
    }

    if(err_status == 0)
    {
        PHWI_ASYNC_PDU_CONTEXT pasync_ctx;
        pasync_ctx = HWI_GET_ASYNC_PDU_CTX(phwi, BE_ULP0_NUM_SLI);

        // No error in ring creation..
        phwi->default_pdu_hdr[BE_ULP0_NUM_SLI].id = be_iscsi_defq_get_id(&pcontext->be_defq[BE_ULP0_NUM_SLI][0]);
        phwi->default_pdu_buf[BE_ULP0_NUM_SLI].id = be_iscsi_defq_get_id(&pcontext->be_defq[BE_ULP0_NUM_SLI][1]);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
            ("Default PDU Rings created with RID %d and RID %d\n",
                    phwi->default_pdu_hdr[BE_ULP0_NUM_SLI].id,
                    phwi->default_pdu_buf[BE_ULP0_NUM_SLI].id));

        pasync_ctx->doorbell_offset = STRUCT_FIELD_OFFSET(PROTECTION_DOMAIN_DBMAP, rxulp0_db);
        pasync_ctx->pci_register_set = PCI_BAR_REGISTER_2_SLI;
        pasync_ctx->doorbell_format = DOORBELL_FORMAT_LIST_SLI;
        pasync_ctx->ulp = BE_ULP0_NUM_SLI;
    }
    else
    {
        if(i == 0)
            err_status = HWI_ERR_DPDU_HDR_RING_CREATE;
        else
            err_status = HWI_ERR_DPDU_DATA_RING_CREATE;
    }

    return err_status;
}

ACIT_UEXACT32
hwi_init_rq_ex(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 err_status = 0, i, def_pdu_ring_sz;
    BESTATUS bestatus;
    ACIT_UEXACT32 ulp;

    for(ulp = 0; ulp < BE_MAX_ULP; ulp++)
    {
        SA_SGL sgl[2] = {0};

        if(p->ulp_mask & (1 << ulp))
        {
            def_pdu_ring_sz = (p->num_async_pdus * sizeof(PHYS_ADDR));
            hwi_build_sa_sgl (pc, HWI_MEM_ASYNC_HEADER0_RING+ulp, &sgl[0]);
            hwi_build_sa_sgl (pc, HWI_MEM_ASYNC_DATA0_RING+ulp, &sgl[1]);

            // The i iterates through header and data
            for(i=0; i<2; i++)
            {
                bestatus = be_iscsi_defq_create_ex (pfo,
                                                 &sgl[i],
                                                 def_pdu_ring_sz, // length
                                                 p->default_pdu_buf_sz,
                                                 &pcontext->phwi_queue[0].be_cq,
                                                 ulp,
                                                 TRUE,
                                                 TRUE, // bind_q_to_ulp,
                                                 &pcontext->be_defq[ulp][i]);

                if (bestatus != BE_SUCCESS)
                {
                    TRACE (DL_ERR, "defq create failed.");
                    err_status = 1;
                    break;
                }
            }

            if(err_status == 0)
            {
                PHWI_ASYNC_PDU_CONTEXT pasync_ctx;
                ACIT_UEXACT16 rid;

                pasync_ctx = HWI_GET_ASYNC_PDU_CTX(phwi, ulp);

                phwi->default_pdu_hdr[ulp].id = be_iscsi_defq_get_id(&pcontext->be_defq[ulp][0]);
                phwi->default_pdu_buf[ulp].id = be_iscsi_defq_get_id(&pcontext->be_defq[ulp][1]);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
                    ("Default PDU Rings created with RID %d and RID %d\n",
                            phwi->default_pdu_hdr[ulp].id, phwi->default_pdu_buf[ulp].id));

                // assume defq data ring info will be the same as defq header info
                be_iscsi_defq_get_exinfo(&pcontext->be_defq[ulp][0],
                                         &rid,
                                         &pasync_ctx->doorbell_offset,
                                         &pasync_ctx->pci_register_set,
                                         &pasync_ctx->doorbell_format,
                                         &pasync_ctx->ulp);

                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                    ("Defq Id %d DBOffset %x RegSet %d DBFormat %d Ulp %d\n",
                    rid,
                    pasync_ctx->doorbell_offset,
                    pasync_ctx->pci_register_set,
                    pasync_ctx->doorbell_format,
                    pasync_ctx->ulp));
            }
            else
            {
                if(i == 0)
                    err_status = HWI_ERR_DPDU_HDR_RING_CREATE;
                else
                    err_status = HWI_ERR_DPDU_DATA_RING_CREATE;
            }
        }
    }

    return err_status;
}

ACIT_UEXACT32
hwi_init_rq(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    ACIT_UEXACT32 err_status = 0, ulp;

    // Initialize the ASYNC PDU context. By this time, the UlpMask
    // has been read from the firmware
    hwi_init_async_pdu_ctx(pc);

    if(pc->params.num_ulp)
    {
        // DUA_TODO: Merge both within _ex routine?
        if(!IS_DUA_MODE(pc))
        {
            err_status = _hwi_init_rq(pc, phwi, pcontext, pfo);
        }
        else
        {
            err_status = hwi_init_rq_ex(pc, phwi, pcontext, pfo);
        }

        if(err_status == 0)
        {
            for(ulp = 0; ulp < BE_MAX_ULP; ulp++)
            {
                if(pc->params.ulp_mask & (1 << ulp))
                {
                    hwi_post_async_buffers(pc, 1, ulp); // post buffer to header ring
                    hwi_post_async_buffers(pc, 0, ulp); // post buffer to data ring
                }
            }
            pcontext->initialized.def_hdrq = 1;
            pcontext->initialized.def_bufq = 1;
        }
    }
        
    return err_status;
}

ACIT_UEXACT32
hwi_init_template_headers(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 err_status = 0, ulp, ulp_offset = 0;
    BESTATUS bestatus;
    IOCTL_COMMON_ADD_TEMPLATE_HEADER_BUFFERS_RESPONSE response; 
    SA_SGL sgl = {0};

    if(p->num_cxn)
    {
        ACIT_UEXACT32 num_pages = p->num_hdr_pages/p->num_ulp;
        OSM_ASSERT(num_pages);

        for(ulp = 0; ulp < BE_MAX_ULP; ulp++)
        {
            if(p->ulp_mask & (1 << ulp))
            {
                memset(&response,0,sizeof(response));

                hwi_build_sa_sgl_by_offset(pc,
                                           HWI_MEM_HDRS,
                                           (num_pages * PAGE_SIZE * ulp_offset), // offset
                                           (num_pages * PAGE_SIZE),       // size
                                           &sgl);

                bestatus = be_function_post_template_header_memory (pfo, &sgl, &response);    
                if (bestatus != BE_SUCCESS)
                {
                    TRACE (DL_ERR, "post templates failed.");
                    err_status = HWI_ERR_TEMPLATE_HEADER_POST;
                    goto error;
                }

                pcontext->initialized.templates = 1;
                ulp_offset++;
            }
        }
    }

error:
    return err_status;
}

ACIT_UEXACT32
hwi_init_sgl_pages(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 err_status = 0;
    BESTATUS bestatus;
    MGMT_STATUS mgmt_status;
    SA_SGL sgl = {0};
	ACIT_UEXACT32 sgl_num_pg_to_post, num_pg_post, desc_index = 0;
    ACIT_UEXACT32 page_offset, num_sge = 0, bits = 0, icd_begin = 0, icd_counter = 0;
    ACIT_UEXACT32 total_posted, total_needed, available, remaining, desc_offset;
    ACIT_UEXACT32 post, ulp, descriptors, use_v1_ioctl = 0;
    
    if(p->icds_per_controller == 0)
    {
        // Do not attempt to post pages if the profile is messed up
        goto error;
    }

    // Set the frag_num bits if 256 SGEs are required for BE2/BE3
    OSM_ASSERT(p->num_sge);
    num_sge = p->num_sge;

    if(IS_ASIC_BE2_OR_BE3(pc))
    {
        if(num_sge != 32)
        {
            while(num_sge != 0x1)
            {
                num_sge = (num_sge >> 1);
                bits++;
            }

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Using %d bits for frag_num in ISCSI_SGL_CRA_V2\n", bits));

            mgmt_status = mgmt_set_fragnum_bits(pc, bits);

            if(mgmt_status != MGMT_STATUS_SUCCESS)
            {
                err_status = HWI_ERR_FRAGNUM_SET;
                TRACE (DL_ERR, "Cannot set fragnum bits\n");
                goto error;
            }
        }
    }
    else
    {
        if(p->num_sge != 32)
        {
            // For SkyHawk when 256 SGEs are required, use v1 of Post SGL IOCTL.
            use_v1_ioctl = 1;
        }
    }

    descriptors = 0;
    desc_index = phwi->frag_mem_index;
    desc_offset = 0; // offset within the descriptor. Needed when the same
                     // descriptor contains both Ulp0 and Ulp1 frag memory
    available = pc->mem_descriptors[desc_index].size_in_bytes / BE_PAGE_SIZE;
    total_posted = 0;

    // Bug 138859 : The ICDs are not distributed per Ulp and therefore the combined
    // number is reported on both Ulp configs. So it's enough to pick from one.
    ulp = ap_get_initial_ulp(pc);

    // Bug 147561: Use page-aligned icd ranges
    icd_begin = pc->fw_config.aligned_icd_start;

    page_offset = (phwi->sge_bytes_per_io * icd_begin)/BE_PAGE_SIZE;
    remaining = (p->icds_per_controller * phwi->sge_bytes_per_io)/BE_PAGE_SIZE;

    while(available != 0)
    {
        // Calculate how many SGLs are needed here.
        // Reason is because ACIT may have allocated memory for FRAG even before
        // determining how many SGLs are supported by the hardware.

        if(remaining < available) 
	        post = remaining;
        else
            post = available;

        hwi_build_sa_sgl_by_offset(pc,
                                   desc_index,
                                   desc_offset * BE_PAGE_SIZE,
                                   remaining * BE_PAGE_SIZE,
                                   &sgl);

        if(use_v1_ioctl == 0)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                    ("Posting SGLPage: page_offset %d desc_offset %d num_pages %d\n",
                    page_offset,
                    desc_offset,
                    post));

            bestatus = be_iscsi_post_sgl_pages (
                                pfo,
                                &sgl,
                                page_offset,
                                post);
        }
        else
        {
            // How many ICDs posted in post number of BE_PAGES?
            ACIT_UEXACT32 icds_posted = (post * BE_PAGE_SIZE)/phwi->sge_bytes_per_io;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,
                    ("Posting SGLPage V1: icd_start %d icd_count %d desc_offset %d num_pages %d\n",
                    icd_begin + icd_counter,
                    icds_posted,
                    desc_offset,
                    post));

            bestatus = be_iscsi_post_sgl_pages_v1 (
                            pfo,
                            &sgl,
                            icd_begin + icd_counter,
                            icds_posted);

            icd_counter += icds_posted;
        }

        OSM_ASSERT(bestatus == BE_SUCCESS);

        if (bestatus != BE_SUCCESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("post sgl failed."));
            err_status = HWI_ERR_SGL_PAGES_POST;
            goto error;
        }

        page_offset += post;
        remaining -= post;
        available -= post;
        desc_offset += post;
        total_posted += post;

        if(available == 0)
        {
            // We've run out of this memory descriptor. Move to the next descriptor
            descriptors++;

            if(descriptors < pc->acit_var_desc_index.run_time.frag)
            {
                desc_index++;
                desc_offset = 0;
                available = pc->mem_descriptors[desc_index].size_in_bytes / BE_PAGE_SIZE;
            }
            else
            {
                // Note! Everything within this block of statement is for debug purpose alone

                ACIT_UEXACT32 total_needed;

                // Bug 138859 : Combined icd_count is reported on both Ulps
                total_needed = p->icds_per_controller;

                total_needed *= phwi->sge_bytes_per_io;
                total_needed /= BE_PAGE_SIZE;

                if((pc->state.init_type != ACIT_INIT_TYPE_CRASHDUMP) &&
                    (pc->state.init_type != ACIT_INIT_TYPE_HIBERNATION))
                {
                    // Remaining has to be 0 AND we should have completed posting
                    // memory for all descriptors
                    if(total_posted != total_needed)
                    {
                        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR , ("posted %d needed %d\n",
                            total_posted, total_needed));
                        OSM_ASSERT(0);
                    }
                }
            }
        }

        if(remaining == 0)
        {
            // No more descriptors needed for this Ulp, break
            // and move to next Ulp
            break;
        }
    }

    pcontext->initialized.sgl_pages = 1;

error:
    return err_status;
}

ACIT_UEXACT32
hwi_init_wrb_rings(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi,
    PHWI_CONTEXT_MEMORY     pcontext,
    PBE_FUNCTION_OBJECT     pfo)
{
    PACIT_PARAMETERS p = &pc->params;
    ACIT_UEXACT32 err_status = 0;
    BESTATUS bestatus;
    MGMT_STATUS mgmt_status;
    SA_SGL sgl = {0};
    ACIT_UEXACT32 wrb_mem_index, offset, size, i;
    PHWI_WRB_CONTEXT pwrb_context = ACIT_NULL;
    ACIT_UEXACT32 ulp = 0;

    // create all the WRB ring objects.

    for (i=0; i<p->num_cxn; i++) {

        wrb_mem_index = 0;
        offset = 0;
        size = 0;
        pwrb_context = &phwi->pwrb_context[i];

        hwi_get_wrb_mem(phwi, i, &wrb_mem_index, &offset, &size);

        hwi_build_sa_sgl_by_offset (pc, 
                                    wrb_mem_index,
                                    offset,
                                    size,
                                    &sgl);

        if(p->num_ulp > 1)
        {
            // Alternate between Ulps at the time of allocating the
            // rings if there is more than 1 Ulp
            ulp = (ulp+1) % p->num_ulp;
        }
        else
        {
            ulp = ap_get_initial_ulp(pc);
        }

        // DUA_TODO: Merge both within _ex routine?
        if(!IS_DUA_MODE(pc))
        {
            bestatus = be_iscsi_wrb_queue_create (pfo,
                                    &sgl,
                                    sa_sgl_get_byte_count (&sgl),
                                    &pcontext->pbe_wrbq[i].be_wrbq);

            if (bestatus != BE_SUCCESS)
            {
                TRACE (DL_ERR, "wrbq create failed.");
                err_status = HWI_ERR_WRB_QUEUE_CREATE;
                goto error;
            }

            // Hardcoded doorbell offset and format
            pwrb_context->doorbell_offset = STRUCT_FIELD_OFFSET(PROTECTION_DOMAIN_DBMAP, txulp0_db);
            pwrb_context->pci_register_set = PCI_BAR_REGISTER_2_SLI;
            pwrb_context->doorbell_format = DOORBELL_FORMAT_RING_SLI;
            pwrb_context->ulp = BE_ULP0_NUM_SLI;
        }
        else
        {
            bestatus = be_iscsi_wrb_queue_create_ex (pfo,
                                    &sgl,
                                    sa_sgl_get_byte_count (&sgl),
                                    ulp,
                                    TRUE,
                                    TRUE, // bind_q_to_ulp
                                    &pcontext->pbe_wrbq[i].be_wrbq);

            if (bestatus != BE_SUCCESS)
            {
                TRACE (DL_ERR, "wrbq create_ex failed.");
                err_status = HWI_ERR_WRB_QUEUE_CREATE;
                goto error;
            }

            be_iscsi_wrbq_get_exinfo(&pcontext->pbe_wrbq[i].be_wrbq,
                                     &pwrb_context->cid,
                                     &pwrb_context->doorbell_offset,
                                     &pwrb_context->pci_register_set,
                                     &pwrb_context->doorbell_format,
                                     &pwrb_context->ulp);
        }

        //
        // This is a table which maps the Connection Resource Index
        // and the CID used by the hardware. CRI is a resource shared
        // with iSCSI and is used so that iSCSI gets a consistent 0
        // based number for managing connection resources.
        //
        pwrb_context->cid = \
           (ACIT_UEXACT16)be_iscsi_wrbq_get_cid(&pcontext->pbe_wrbq[i].be_wrbq);

        pcontext->initialized.num_wrbq = i + 1;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
           ("Cri %d iSCSI WRB Ring Id 0x%x DBOffset %x RegSet %d DBFormat %d Ulp %d\n",
                i, pwrb_context->cid,
                pwrb_context->doorbell_offset,
                pwrb_context->pci_register_set,
                pwrb_context->doorbell_format,
                pwrb_context->ulp));
    }

    // Initialize the table that maps the CID to a CRI
    err_status = hwi_init_cid_map(pc, p->num_cxn); // Number of elements

    if(err_status != ACIT_STATUS_SUCCESS)
    {
        err_status = HWI_ERR_CID_MAP_TABLE_CREATE;
        goto error;
    }

error:
    return err_status;
}

#if 0
hwi_init_msi_messages()
{
    // inform the chip about the number of msi-x interrupts we need
    // Fixme!!! Do not call this for non-MSIx mode
    if(pc->params.num_cq)
    {
        bestatus = be_cev_modify_msi_messages(pfo, 
                                              pc->params.num_cq,
                                              ACIT_NULL, ACIT_NULL, ACIT_NULL);

        if(bestatus != BE_SUCCESS)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("ModifyMSIMessages failed msix %d status 0x%x\n", 
                                    pc->params.num_cq, bestatus));
            OSM_ASSERT(0);
        }
    }
}
#endif

ACIT_VOID
hwi_init_vase_memory(
    PACIT_CONTROLLER_WS     pc,
    PHWI_CONTROLLER_WS      phwi)
{
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,"psgl_handle_base",phwi->psgl_handle_base);
//    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,"pasync_buff_base",phwi->pasync_buff_base);
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,"wrb_base",phwi->wrb_base[0]);
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,"pmcc_wrb_base",phwi->pmcc_wrb_base);
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,"CQ Base",
                    (ACIT_UEXACT64)pc->mem_descriptors[phwi->cq_mem_index].virtual_address);
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,"EQ Base",
                    (ACIT_UEXACT64)pc->mem_descriptors[HWI_MEM_EQ].virtual_address);
    DBG_LOG_OP_PARAMETER(&pc->vase_master_block,"MCC CQ Base",
                    (ACIT_UEXACT64)pc->mem_descriptors[HWI_MEM_MCC_CQ].virtual_address);
}
